#!/usr/bin/env python
""" 
Collection of functions to filter audio signal in time and frequency
domains. 
"""
#
# Authors:  Juan Sebastian ULLOA <lisofomia@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

# =============================================================================
# Load the modules
# =============================================================================
# Import external modules
import numpy as np
import matplotlib.pyplot as plt
from scipy.signal import (
    sosfiltfilt, convolve, iirfilter, get_window, 
    kaiserord, firwin, lfilter
    )
from skimage import filters

# Import internal modules 
from maad.util import plot2d

#%%
# =============================================================================
# public functions
# =============================================================================
def select_bandwidth (x, fs, fcut, forder, fname ='butter', ftype='bandpass', 
                    rp=None, rs=None):
    """
    Use a lowpass, highpass, bandpass or bandstop filter to process a 1d signal with an iir filter.
        
    Parameters
    ----------
    x : array_like
        1d vector of scalars to be filtered
    fs : scalar
        sampling frequency   
    fcut : array_like
        A scalar or length-2 sequence giving the critical frequencies.             
    forder : int
        The order of the filter.
    fname : {`butter`, 'cheby1`, `cheby2`, `ellip`, `bessel`}, optional, default
        is 'butter'   
        The type of IIR filter to design:
        - Butterworth : `butter`
        - Chebyshev I : 'cheby1`
        - Chebyshev II : `cheby2`
        - Cauer/elliptic: `ellip`
        - Bessel/Thomson: `bessel`   
    ftype : {`bandpass`, `lowpass`, `highpass`, `bandstop`}, optional, default 
        is `bandpass`
        The type of filter.
    rp : float, optional
        For Chebyshev and elliptic filters, provides the maximum ripple in 
        the passband. (dB)   
    rs : float, optional
        For Chebyshev and elliptic filters, provides the minimum attenuation 
        in the stop band. (dB)           
            
    Returns
    -------
    y : array_like
        The filtered output with the same shape and phase as x
        
    See Also
    --------
    fir_filter1d
        Lowpass, highpass, bandpass or bandstop a 1d signal with an Fir filter
    
    Examples
    --------
    >>> import maad

    Load and display the spectrogram of a sound waveform
    
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> Sxx_power,tn,fn,_ = maad.sound.spectrogram(w,fs)
    >>> Sxx_dB = maad.spl.power2dBSPL(Sxx_power, gain=42) # convert into dB SPL
    >>> fig_kwargs = {'vmax': Sxx_dB.max(),\
                    'vmin':0,\
                    'extent':(tn[0], tn[-1], fn[0], fn[-1]),\
                    'figsize':(4,13),\
                    'title':'Power spectrogram density (PSD)',\
                    'xlabel':'Time [sec]',\
                    'ylabel':'Frequency [Hz]',\
                    }
    >>> ax, fig = maad.util.plot2d(Sxx_dB, **fig_kwargs)
    >>> print('The energy in the spectrogram is {} dB SPL'.format(maad.util.add_dB(maad.util.add_dB(Sxx_dB))))
    The energy in the spectrogram is 90.68996875800728 dB SPL
    
    Filter the waveform : keep the bandwidth between 6-10kHz
    
    >>> w_filtered = maad.sound.select_bandwidth(w,fs,fcut=(6000,10000), forder=5, fname ='butter', ftype='bandpass')
    >>> Sxx_power_filtered,tn,fn,_ = maad.sound.spectrogram(w_filtered,fs)
    >>> Sxx_dB_filtered = maad.spl.power2dBSPL(Sxx_power_filtered, gain=42) # convert into dB SPL
    >>> ax, fig = maad.util.plot2d(Sxx_dB_filtered, **fig_kwargs)
    >>> print('The energy in the spectrogram is {} dB SPL'.format(maad.util.add_dB(maad.util.add_dB(Sxx_dB_filtered))))
    The energy in the spectrogram is 72.75045485363799 dB SPL

    """
    sos = iirfilter(N=forder, Wn=np.asarray(fcut)/(fs/2), btype=ftype,ftype=fname, rp=rp, 
                    rs=rs, output='sos')
    # use sosfiltfilt insteasd of sosfilt to keep the phase of y matches x
    y = sosfiltfilt(sos, x)
    return y

#%%
def fir_filter(x, kernel, axis=0):
    """
    Filter a signal using a 1d finite impulse response filter.
    
    This function uses a digital filter based on convolution of 1d kernel over a vector 
    or along an axis of a matrix.
    
    Parameters
    ----------
    x : array_like
        1d vector or 2d matrix of scalars to be filtered
    kernel : array_like or tuple
        Pass directly the kernel (1d vector of scalars) 
        Or pass the arguments in a tuple to create a kernel. Arguments are:   
        - window : string, float, or tuple. The type of window to create. 
        boxcar, triang, blackman, hamming, hann, bartlett, flattop,
        parzen, bohman, blackmanharris, nuttall, barthann, 
        - (kaiser, beta), 
        - (gaussian, standard deviation), 
        - (general_gaussian, power, width), 
        - (slepian, width), 
        - (dpss, normalized half-bandwidth), 
        - (chebwin, attenuation), 
        - (exponential, decay scale), 
        - (tukey, taper fraction)
        - N : length of the kernel
        Examples:
        - kernel = ('boxcar', 9)
        - kernel = (('gaussian', 0.5), 5)
        - kernel = [1 3 5 7 5 3 1] 
    axis : int
        Determine along which axis is performed the filtering in case of 2d matrix
        axis = 0 : vertical
        axis = 1 : horizontal
    
    Returns
    -------
    y : array_like
        The filtered output with the same shape and phase as x
        
    See Also
    --------
    select_bandwidth
        Lowpass, highpass, bandpass or bandstop a 1d signal with an iir filter
    
    Examples
    --------  
    >>> import maad
    
    Load and display the spectrogram of a sound waveform
    
    >>> w, fs = maad.sound.load('../data/cold_forest_daylight.wav') 
    >>> Sxx_power, tn, fn, ext = maad.sound.spectrogram(w,fs)
    >>> Lxx = maad.spl.power2dBSPL(Sxx_power, gain=42) # convert into dB SPL
    >>> fig_kwargs = {'vmax': Lxx.max(),\
                    'vmin':0,\
                    'extent': ext,\
                    'figsize': (4,13),\
                    'title': 'Power spectrogram density (PSD)',\
                    'xlabel': 'Time [sec]',\
                    'ylabel': 'Frequency [Hz]',\
                    }
    >>> ax, fig = maad.util.plot2d(Lxx,**fig_kwargs)

    >>> print('The energy in the spectrogram is {} dB SPL'.format(maad.util.add_dB(maad.util.add_dB(Lxx))))
    The energy in the spectrogram is 90.68996875800728 dB SPL
    
    Smooth the waveform (lowpass)
    
    >>> w_filtered = maad.sound.fir_filter(w, kernel=(('gaussian', 2), 5))
    >>> Sxx_power_filtered,tn,fn,_ = maad.sound.spectrogram(w_filtered,fs)
    >>> Lxx_filtered = maad.spl.power2dBSPL(Sxx_power_filtered, gain=42) # convert into dB SPL
    >>> ax, fig = maad.util.plot2d(Lxx_filtered,**fig_kwargs)

    >>> print('The energy in the spectrogram is {} dB SPL'.format(maad.util.add_dB(maad.util.add_dB(Lxx_filtered))))
    The energy in the spectrogram is 89.51776241848223 dB SPL
    
    Smooth the spectrogram, frequency by frequency (blurr)
    
    >>> Lxx_blurr = maad.sound.fir_filter(Lxx, kernel=(('gaussian', 1), 5), axis=1)
    >>> ax, fig = maad.util.plot2d(Lxx_blurr,**fig_kwargs)

    >>> print('The energy in the spectrogram is {} dB SPL'.format(maad.util.add_dB(maad.util.add_dB(Lxx_blurr))))
    The energy in the spectrogram is 88.83611122906812 dB SPL
    
    """
    if isinstance(kernel,tuple) :
        if len(kernel) ==1 :
            win = get_window(kernel[0])
        if len(kernel) ==2 :
            win = get_window(kernel[0], kernel[1])
    elif isinstance(kernel,list) or isinstance(kernel,np.ndarray):
        win = kernel

    if x.ndim == 1:
        # Trick to avoid strange values at the beginning and end of the convolve
        # signal. Add mirror values
        x = np.insert(x, 0, np.flipud(x[0:len(win)//2]), axis=0) 
        x = np.insert(x, len(x), np.flipud(x[-(len(win)//2):]), axis=0) 
        # convolve and normalized the result by the sum of the kernel
        y = convolve(x, win, mode='same') / sum(win)
        y = y[len(win)//2:-(len(win)//2)]
    elif x.ndim == 2:
        
        if axis ==1 : x = x.transpose()
        
        # Trick to avoid strange values at the beginning and end of the convolve
        # signal. Add mirror values
        x = np.insert(x, 0, np.flipud(x[0:len(win)//2]), axis=0) 
        x = np.insert(x,  x.shape[0], np.flipud(x[-(len(win)//2):]), axis=0) 
        y = np.zeros(x.shape)
        for i in np.arange (x.shape[1]):
            y[:,i] = convolve(x[:,i], win, mode='same') / sum(win)
        y = y[len(win)//2:-(len(win)//2),:]
        
        if axis ==1 :y = y.transpose()

    return y

#%%
def sinc(s, cutoff, fs, atten=80, transition_bw=0.05, bandpass=True):
    """
    Filter 1D signal with a Kaiser-windowed filter.
    
    Parameters
    ----------
    s : ndarray
        input 1D signal
    cutoff : ndarray
        upper and lower frequencies (min_f, max_f) in Hertz
    atten : float 
        attenuation in dB
    transition_bw : float
        transition bandwidth in percent default 5% of total band
    bandpass : bool
        bandpass (True) or bandreject (False) filter, default is bandpass
    
    Returns
    -------
    s_filt : array 
        signal filtered
    
    Examples
    --------
    >>> import maad
    >>> s, fs = maad.sound.load('../data/spinetail.wav')
    >>> import numpy as np
    >>> tn = np.arange(0,len(s))/fs
    >>> s_filt_7_12kHz = maad.sound.sinc(s, cutoff=[7000,12000], fs=fs, atten=80, transition_bw=0.8)
    >>> s_filt_4_8kHz = maad.sound.sinc(s, cutoff=[4500,8000], fs=fs, atten=80, transition_bw=0.8)
    >>> import matplotlib.pyplot as plt
    >>> fig, (ax0, ax1,ax2) = plt.subplots(3,1, sharex=True, squeeze=True)
    >>> ax0, fig = maad.util.plot1d(tn,s,ax=ax0, figtitle='original')
    >>> ax1, fig = maad.util.plot1d(tn,s_filt_7_12kHz,ax=ax1, figtitle='Kaiser-windowed filter 7-12kHz')
    >>> a2 , fig = maad.util.plot1d(tn,s_filt_4_8kHz,ax=ax2, figtitle='Kaiser-windowed filter 4.5-8kHz')
    >>> fig.tight_layout()
    
    """
    width = (cutoff[1] - cutoff[0]) * transition_bw
    numtaps, beta = kaiserord(atten, width/(0.5*fs))
    np.ceil(numtaps-1) // 2 * 2 + 1  # round to nearest odd to have Type I filter
    taps = firwin(
        numtaps, 
        cutoff, 
        window=('kaiser', beta), 
        scale=False, 
        fs=fs, 
        pass_zero=not(bandpass)
        )
    s_filt = lfilter(taps, 1, s)
    return s_filt

#%%
def smooth (Sxx, std: float=1, verbose=False, display = False, savefig=None, **kwargs): 
    """ 
    Smooth a spectrogram with a gaussian filter.
    
    Parameters 
    ---------- 
    Sxx : 2d ndarray of scalars 
        Spectrogram (or image) 
    
    std : scalar, optional, default is 1 
        Standard deviation of the gaussian kernel used to smooth the spectrogram.
        The larger is the number, the smoother will be the image and the longer 
        it takes. Standard values should fall between 0.5 and 3.
    
    verbose : boolean, optional, default is False 
        Print messages
    
    display : boolean, optional, default is False 
        Display the signal if True 
        
    savefig : string, optional, default is None 
        Root filename (with full path) is required to save the figures. Postfix 
        is added to the root filename. 
        
    kwargs, optional. This parameter is used by plt.plot and savefig functions 
        
        - savefilename : str, optional, default :'_spectro_after_noise_subtraction.png' 
            Postfix of the figure filename 
        
        - extent : list of scalars [left, right, bottom, top], optional, default: None
            The location, in data-coordinates, of the lower-left and
            upper-right corners. If `None`, the image is positioned such that
            the pixel centers fall on zero-based (row, column) indices.
        
        - title : string, optional, default : 'Spectrogram' 
            title of the figure 
        
        - xlabel : string, optional, default : 'Time [s]' 
            label of the horizontal axis 
        
        - ylabel : string, optional, default : 'Amplitude [AU]' 
            label of the vertical axis 
        
        - cmap : string or Colormap object, optional, default is 'gray' 
            See https://matplotlib.org/examples/color/colormaps_reference.html 
            in order to get all the  existing colormaps 
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic',  
            'viridis'... 
        
        - vmin, vmax : scalar, optional, default: None 
            `vmin` and `vmax` are used in conjunction with norm to normalize 
            luminance data.  Note if you pass a `norm` instance, your 
            settings for `vmin` and `vmax` will be ignored. 
        
        - extent : scalars (left, right, bottom, top), optional, default: None 
            The location, in data-coordinates, of the lower-left and 
            upper-right corners. If `None`, the image is positioned such that 
            the pixel centers fall on zero-based (row, column) indices. 
        
        - dpi : integer, optional, default is 96 
            Dot per inch.  
            For printed version, choose high dpi (i.e. dpi=300) => slow 
            For screen version, choose low dpi (i.e. dpi=96) => fast 
        
        - format : string, optional, default is 'png' 
            Format to save the figure 
            
        ... and more, see matplotlib    
        
    Returns 
    ------- 
    im_out: smothed or blurred image  
    
    Examples
    --------
    >>> import maad
    
    Load audio recording and convert it into spectrogram
    
    >>> s, fs = maad.sound.load('../data/cold_forest_daylight.wav')
    >>> Sxx,tn,fn,ext = maad.sound.spectrogram (s, fs, tcrop=(5,10), fcrop=(0,10000))   
    
    Convert linear spectrogram into dB
    
    >>> Sxx_dB = maad.util.power2dB(Sxx) +96
    
    Smooth the spectrogram
    
    >>> Sxx_dB_std05 = maad.sound.smooth(Sxx_dB, std=0.5)
    >>> Sxx_dB_std10 = maad.sound.smooth(Sxx_dB, std=1)
    >>> Sxx_dB_std15 = maad.sound.smooth(Sxx_dB, std=1.5)
    
    Plot spectrograms
    
    >>> import matplotlib.pyplot as plt 
    >>> fig, (ax1, ax2, ax3, ax4) = plt.subplots(4, 1)
    >>> ax1, fig = maad.util.plot2d(Sxx_dB, ax=ax1, extent=ext, title='original', vmin=10, vmax=70)
    >>> ax2, fig = maad.util.plot2d(Sxx_dB_std05, ax=ax2, extent=ext, title='smooth (std=0.5)', vmin=10, vmax=70)
    >>> ax3, fig = maad.util.plot2d(Sxx_dB_std10, ax=ax3, extent=ext, title='smooth (std=1)', vmin=10, vmax=70)
    >>> ax4, fig = maad.util.plot2d(Sxx_dB_std15, ax=ax4, extent=ext, title='smooth (std=1.5)', vmin=10, vmax=70)
    >>> fig.set_size_inches(7,9)
    >>> fig.tight_layout() 
    
    """ 
    
    if verbose:
        print(72 * '_') 
        print('Smooth the image with a gaussian filter (std = %.1f)' %std) 
    
    # use scikit image (faster than scipy) 
    Sxx_out = filters.gaussian(Sxx,std) 
    
    # Display 
    if display :  
        ylabel =kwargs.pop('ylabel','Frequency [Hz]') 
        xlabel =kwargs.pop('xlabel','Time [sec]')   
        cmap   =kwargs.pop('cmap','gray')  
        vmin=kwargs.pop('vmin',np.percentile(Sxx_out,0.1)) 
        vmax=kwargs.pop('vmax',np.percentile(Sxx_out,99.9)) 
        extent=kwargs.pop('extent',None)
            
        if extent is None : 
            ylabel = "pseudofrequency [points]"
            xlabel = "pseudotime [points]"
        
        fig, (ax1, ax2) = plt.subplots(2, 1)
        
        plot2d (Sxx, ax=ax1, extent=extent, 
                title=('Orignal Spectrogram'),  
                ylabel = ylabel, xlabel = xlabel, 
                vmin=vmin, vmax=vmax, 
                cmap=cmap, **kwargs)
        
        plot2d (Sxx_out, ax=ax2, extent=extent, 
                title='Blurred Spectrogram (std='+str(std)+')',  
                ylabel = ylabel, xlabel = xlabel, 
                vmin=vmin, vmax=vmax, 
                cmap=cmap, **kwargs) 
        
        # SAVE FIGURE 
        if savefig is not None :  
            dpi   =kwargs.pop('dpi',96) 
            format=kwargs.pop('format','png') 
            filename=kwargs.pop('filename','_spectro_blurr')              
            filename = savefig+filename+'.'+format 
            print('\n''save figure : %s' %filename) 
            fig.savefig(filename, bbox_inches='tight', dpi=dpi, format=format, 
                        **kwargs)    
    return Sxx_out 

if __name__ == "__main__":
    import doctest
    doctest.testmod()


# -*- coding: utf-8 -*-
""" Multiresolution Analysis of Acoustic Diversity Scikit (Toolbox for SciPy)

scikit-maad is a modular toolbox to analyze ecoacoustics datasets in Python 3. 
This package was designed to bring flexibility to find regions of interest,
and to compute acoustic features in audio recordings. This workflow opens 
the possibility to use powerfull machine learning algorithms through 
scikit-learn, allowing to identify key patterns in all kind of soundscapes.

Subpackages
-----------

acoustics
    Collection of functions to describe the physics of acoustic waves such as
    sound pressure level, attenuation laws or propagation - :ref:`maad.acoustics`

features
    Compute descriptors to characterize sounds - :ref:`maad.features`

rois
    Find regions of interest in 1D and 2D signals - :ref:`maad.rois`

sound
    Load, preprocess and transform (e.g. stft) audio signals - :ref:`maad.sound`

spl
    Compute Sound Pressure Level (SPL) - :ref:`maad.spl`
    
util
    Miscelaneous and useful set of tools used in the audio analysis framework - :ref:`maad.util`    
"""
from .version import __version__

# from . import spl
# from . import features
# from . import rois
# from . import sound
# from . import util



#!/usr/bin/env python3 
# -*- coding: utf-8 -*- 
""" 
Ensemble of functions to compute acoustic descriptors from 2D spectrograms.
""" 
#
# Authors:  Juan Sebastian ULLOA <lisofomia@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>        
#
# License: New BSD License

#%%
#***************************************************************************
# -------------------       Load modules         ---------------------------
#***************************************************************************
# Import external modules

from __future__ import print_function 
import numpy as np 
import pandas as pd 
from scipy import ndimage as ndi
import itertools as it 
import matplotlib.pyplot as plt 
from skimage import transform, measure 
from scipy import ndimage 

# Import internal modules
from maad.util import format_features, overlay_rois, plot_shape, overlay_centroid
from maad.rois import rois_to_imblobs
from maad.sound import spectrogram 
 
#%%
#============================================================================ 
# PRIVATE FUNCTIONS 
#============================================================================ 
def _gabor_kernel_nodc(frequency, theta=0, bandwidth=1, gamma=1, 
                      n_stds=3, offset=0): 
    """ 
    Computes complex 2D Gabor filter kernel with no DC offset. 
         
    Gabor kernel is a Gaussian kernel modulated by a complex harmonic function. 
    Harmonic function consists of an imaginary sine function and a real 
    cosine function. Spatial frequency is inversely proportional to the 
    wavelength of the harmonic and to the standard deviation of a Gaussian 
    kernel. The bandwidth is also inversely proportional to the standard 
    deviation. 
     
    Parameters 
    ---------- 
    frequency : float 
        Spatial frequency of the harmonic function. Specified in pixels. 
    theta : float, optional 
        Orientation in radians. If 0, the harmonic is in the x-direction. 
    bandwidth : float, optional 
        The bandwidth captured by the filter. For fixed bandwidth, `sigma_x` 
        and `sigma_y` will decrease with increasing frequency. This value is 
        ignored if `sigma_x` and `sigma_y` are set by the user. 
    gamma : float, optional 
        gamma changes the aspect ratio (ellipsoidal) of the gabor filter.  
        By default, gamma=1 which means no aspect ratio (circle) 
         
        if gamma>1, the filter is larger (x-dir) 
         
        if gamma<1, the filter is higher (y-dir) 
         
        This value is ignored if `sigma_x` and `sigma_y` are set by the user. 
    sigma_x, sigma_y : float, optional 
        Standard deviation in x- and y-directions. These directions apply to 
        the kernel before rotation. If `theta = pi/2`, then the kernel is 
        rotated 90 degrees so that `sigma_x` controls the vertical direction.    
    n_stds : scalar, optional 
        The linear size of the kernel is n_stds (3 by default) standard 
        deviations 
    offset : float, optional 
        Phase offset of harmonic function in radians. 
     
    Returns 
    ------- 
    g_nodc : complex 2d array 
        A complex gabor kernel with no DC offset 
     
    Notes 
    ----- 
    This function is a modification of the gabor_kernel function of scikit-image. 
     
    References 
    ---------- 
    .. [1] http://en.wikipedia.org/wiki/Gabor_filter 
    .. [2] http://mplab.ucsd.edu/tutorials/gabor.pdf 
    .. [3] http://www.cs.rug.nl/~imaging/simplecell.html 
     
    """ 
     
     # set gaussian parameters 
    b = bandwidth 
    sigma_pref = 1.0 / np.pi * np.sqrt(np.log(2) / 2.0) * (2.0 ** b + 1) / (2.0 ** b - 1) 
    sigma_y = sigma_pref / frequency 
    sigma_x = sigma_y / gamma 
    # meshgrid 
    x0 = np.ceil(max(np.abs(n_stds * sigma_x * np.cos(theta)), 
                     np.abs(n_stds * sigma_y * np.sin(theta)), 1)) 
    y0 = np.ceil(max(np.abs(n_stds * sigma_y * np.cos(theta)), 
                     np.abs(n_stds * sigma_x * np.sin(theta)), 1)) 
    y, x = np.mgrid[-y0:y0 + 1, -x0:x0 + 1] 
    # rotation matrix 
    rotx = x * np.cos(theta) + y * np.sin(theta) 
    roty = -x * np.sin(theta) + y * np.cos(theta) 
    # combine gambor and  
    g = np.zeros(y.shape, dtype=complex) 
    g[:] = np.exp(-0.5 * (rotx ** 2 / sigma_x ** 2 + roty ** 2 / sigma_y ** 2)) 
    g /= 2 * np.pi * sigma_x * sigma_y # gaussian envelope 
    oscil = np.exp(1j * (2 * np.pi * frequency * rotx + offset)) # harmonic / oscilatory function 
    g_dc = g*oscil 
    # remove dc component by subtracting the envelope weighted by K 
    K = np.sum(g_dc)/np.sum(g) 
    g_nodc = g_dc - K*g 
  
    return g_nodc 
 
#%%
def _plot_filter_bank(kernels, frequency, ntheta): 
    """ 
    Display filter bank 
     
    Parameters 
    ---------- 
    kernels: list 
        List of kernels from filter_bank_2d_nodc() 
         
    frequency: 1d ndarray of scalars 
        Spatial frequencies used to built the Gabor filters. Values should be 
        in [0;1] 
     
    ntheta: int 
        Number of angular steps between 0° to 90° 
     
    Returns 
    ------- 
    fig : Figure 
        The Figure instance  
    ax : Axis 
        The Axis instance 
    """ 
 
    ntheta = ntheta 
    nfreq = len(frequency) 
     
    # get maximum size 
    aux = list() 
    for kernel in kernels: 
        aux.append(max(kernel[0].shape)) 
    max_size = np.max(aux) 
     
    # plot kernels 
    fig, ax = plt.subplots(nfreq, ntheta) 
    ax = ax.transpose() 
    ax = ax.ravel() 
    for idx, k in enumerate(kernels): 
        kernel = np.real(k[0]) 
        ki, kj = kernel.shape 
        ci, cj = int(max_size/2 - ki/2), int(max_size/2 - kj/2) 
        canvas = np.zeros((max_size,max_size)) 
        canvas[ci:ci+ki,cj:cj+kj] = canvas[ci:ci+ki,cj:cj+kj] + kernel 
        ax[idx].imshow(canvas) 
        ax[idx].axis('off') 
    plt.tight_layout() 
    plt.subplots_adjust(wspace=0.02, hspace=0.02) 
    return ax, fig 
 
#%% 
def _plot_filter_results(im_ref, im_list, kernels, params, m, n): 
    """ 
    Display the resulting spectrograms after filtering with gabor filters 
     
    Parameters 
    ---------- 
    im_ref : 2D array 
        Reference image 
    im_list : list 
        List of filtered images 
    kernels: list 
        List of kernels from filter_bank_2d_nodc() 
    m: int 
        number of columns 
    n: int 
        number of rows 
         
    Returns 
    ------- 
    fig : Figure 
        The Figure instance  
    ax : Axis 
        The Axis instance 
    """     
         
    ncols = m  
    nrows = n 
     
    fig, axes = plt.subplots(nrows=nrows, ncols=ncols, figsize=(15, 5)) 
    plt.gray() 
    fig.suptitle('Image responses for Gabor filter kernels', fontsize=12) 
    axes[0][0].axis('off') 
     
    # Plot original images 
    axes[0][1].imshow(im_ref, origin='lower') 
    axes[0][1].set_title('spectrogram', fontsize=9) 
    axes[0][1].axis('off') 
    plt.tight_layout 
     
    params_label = [] 
    for param in params: 
        params_label.append('theta=%d,\nf=%.2f' % (param[1] * 180 / np.pi, param[0])) 
     
    ii = 0 
    for ax_row in axes[1:]:  
        plotGabor = True 
        for ax in ax_row: 
            if plotGabor == True: 
                # Plot Gabor kernel 
                print(params_label[ii]) 
                ax.imshow(np.real(kernels[ii]), interpolation='nearest') 
                ax.set_ylabel(params_label[ii], fontsize=7) 
                ax.set_xticks([]) 
                ax.set_yticks([]) 
                plotGabor = False 
            else: 
                im_filtered = im_list[ii] 
                ax.imshow(im_filtered, origin='lower') 
                ax.axis('off') 
                plotGabor = True 
                ii=ii+1      
         
    plt.show() 
    return ax, fig 
 
#%%
def _filter_mag(im, kernel): 
    """ 
    Normalizes the image and computes im and real part of filter response using  
    the complex kernel and the modulus operation 
     
    Parameters 
    ---------- 
    im: 2D array 
        Input image to process  
    kernel: 2D array 
        Complex kernel or filter 
             
    Returns 
    ------- 
    im_out: Modulus operand on filtered image 
 
    """     
     
    im = (im - im.mean()) / im.std() 
    im_out = np.sqrt(ndi.convolve(im, np.real(kernel), mode='reflect')**2 + 
                   ndi.convolve(im, np.imag(kernel), mode='reflect')**2) 
    return im_out 
 
#%%
def _params_to_df(params_filter_bank, npyr): 
    """ 
    Organises parameters used in multiresolution analysis into a dataframe 
     
    Parameters 
    ---------- 
    params_filter_bank: numpy array 
        Output of parameters from function filter_bank_2d_nodc 
    npyr: int 
        Number of pyramids used in multiresolution analysis 
             
    Returns 
    ------- 
    params_multires: pandas DataFrame 
        Ordered parameters for each shape feature used 
    """     
 
    params = np.asarray(params_filter_bank) 
    orient = params[:,0] * 180 / np.pi 
    orient = orient.tolist() * npyr 
    pyr_level = np.sort(np.arange(npyr).tolist()*len(params))+1 
    freq = params[:,1].tolist() * npyr 
    nparams = len(params) * npyr 
    params_multires = np.zeros(nparams, dtype={'names':('theta', 'freq', 'pyr_level','scale'), 
                                               'formats':('f8', 'f8', 'f8','f8')}) 
    params_multires['theta'] = orient 
    params_multires['freq'] = freq 
    params_multires['scale'] = 1 / np.asarray(freq) 
    params_multires['pyr_level'] = pyr_level 
    params_multires = pd.DataFrame(params_multires) 
 
    return params_multires 
 
#%%
#============================================================================ 
# Public functions 
#============================================================================ 
def filter_multires(Sxx, kernels, npyr=4, rescale=True): 
    """ 
    Computes 2D wavelet coefficients at multiple scales using Gaussian pyramid  
    transformation to downscale the input spectrogram. 
     
    Parameters 
    ---------- 
    Sxx: list of 2D arrays 
        List of input spectrograms to filter 
    kernels: list of 2D arrays 
        List of 2D kernels or filters 
    npyr: int, optional 
        Number of pyramids to compute. Default is 4. 
    rescale: boolean, optional 
        Indicates if the reduced images should be rescaled. Default is True. 
             
    Returns 
    ------- 
    Sxx_out: list of 2D arrays 
        List of spectrograms filtered by each 2D kernel 
     
    Examples 
    -------- 
    >>> import maad
    >>> from maad.sound import load, spectrogram 
    >>> from maad.features import filter_bank_2d_nodc, filter_multires 
    >>> from maad import util
    >>> s, fs = load('../data/spinetail.wav') 
    >>> Sxx, dt, df, ext = spectrogram(s, fs) 
    >>> Sxx_db = util.power2dB(Sxx, db_range=80) + 80
    >>> ax, fig = util.plot2d(Sxx_db, **{'extent':ext})
    >>> params, kernels = filter_bank_2d_nodc(frequency=(0.5, 0.25), ntheta=2,gamma=2) 
    >>> Sxx_out = filter_multires(Sxx, kernels, npyr=2)
    
    Plot one of the resulting spectrograms.
    
    >>> ax, fig = util.plot2d(Sxx_out[5], **{'extent':ext})
     
    """     
    # Downscale image using gaussian pyramid  
    if npyr<2:  
        print('Warning: npyr should be int and larger than 2 for multiresolution')  
        im_pyr = tuple(transform.pyramid_gaussian(Sxx, downscale=2,   
                                                  max_layer=1)) 
    else:      
        im_pyr = tuple(transform.pyramid_gaussian(Sxx, downscale=2,   
                                                  max_layer=npyr-1)) 
  
    # filter 2d array at multiple resolutions using gabor kernels  
    im_filt=[]  
    for im in im_pyr:  # for each pyramid  
        for kernel, param in kernels:  # for each kernel  
            im_filt.append(_filter_mag(im, kernel))  #  magnitude response of filter  
      
    # Rescale image using gaussian pyramid  
    if rescale:  
        dims_raw = Sxx.shape  
        Sxx_out=[]  
        for im in im_filt:  
            ratio = np.array(dims_raw)/np.array(im.shape)  
            if ratio[0] > 1:  
                im = transform.rescale(im, scale = ratio, mode='reflect',  
                                       anti_aliasing=True) 
            else:  
                pass  
            Sxx_out.append(im)  
    else:  
        pass  
  
    return Sxx_out  
 
#%%
def filter_bank_2d_nodc(frequency, ntheta, bandwidth=1, gamma=2, display=False, **kwargs): 
    r""" 
    Build an ensemble of complex 2D Gabor filters with no DC offset. 
    
    A Gabor kernel is a Gaussian kernel modulated by a complex harmonic function. 
    Harmonic function consists of an imaginary sine function and a real 
    cosine function. Spatial frequency is inversely proportional to the 
    wavelength of the harmonic and to the standard deviation of a Gaussian 
    kernel. The bandwidth is also inversely proportional to the standard 
    deviation [1]_. These filters have been used to classify textures in images [4]_ and 
    can be applied to characterize spectrograms to classify sounds [5]_.
     
    Parameters 
    ---------- 
    frequency: 1d ndarray of scalars 
        Spatial frequencies used to built the Gabor filters. Values should be 
        in the range [0;1] 
     
    ntheta: int 
        Number of angular steps between 0° to 90° 
     
    bandwidth: scalar, optional, default is 1 
        Spatial-frequency bandwidth of the filter. This parameters affect 
        the resolution of the filters. Lower bandwidth result in sharper 
        in filters with more details. 
     
    gamma: scalar, optional, default is 1 
        Gaussian window that modulates the continuous sine. 
        For ``gamma = 1``, the result is the same gaussian window in x and y direction (circle). 
        For ``gamma <1``, the result is an increased elongation of the filter size in the y direction (elipsoid). 
        For ``gamma >1``, the result is a reduction of the filter size in the y direction (elipsoid). 
     
    display: bool 
        Display a visualization of the filter bank. Default is False. 
 
    Returns 
    ------- 
    params: 2d structured array 
         Parameters used to calculate 2D gabor kernels.  
         Params array has 4 fields (theta, freq, bandwidth, gamma) 
         This can be useful to interpret the result of the filtering process. 
             
    kernels: 2d ndarray of complex values 
         Complex Gabor kernels 
    
    References 
    ---------- 
    .. [1] http://en.wikipedia.org/wiki/Gabor_filter 
    .. [2] http://mplab.ucsd.edu/tutorials/gabor.pdf 
    .. [3] http://www.cs.rug.nl/~imaging/simplecell.html 
    .. [4] Sifre, L., & Mallat, S. (2013). Rotation, scaling and deformation invariant scattering for texture discrimination. Computer Vision and Pattern Recognition (CVPR), 2013 IEEE Conference On, 1233–1240. `DOI: 10.1109/CVPR.2013.163 <https://doi.org/10.1109/CVPR.2013.163>`_
    .. [5] Ulloa, J. S., Aubin, T., Llusia, D., Bouveyron, C., & Sueur, J. (2018). Estimating animal acoustic diversity in tropical environments using unsupervised multiresolution analysis. Ecological Indicators, 90, 346–355. `DOI: 10.1016/j.ecolind.2018.03.026 <https://doi.org/10.1016/j.ecolind.2018.03.026>`_

    
    See also
    --------
    filter_multires, opt_shape_presets, shape_features
    
    Examples 
    -------- 
    >>> import maad

    Build a filter bank using predefined parameters with the function maad.features.opt_shape_presets.
 
    >>> from maad.features import filter_bank_2d_nodc, opt_shape_presets 
    >>> opt = opt_shape_presets(resolution='med') 
    >>> params, kernels = filter_bank_2d_nodc(opt['frequency'], opt['ntheta'], opt['bandwidth'], opt['gamma'], display=True) 
 
    Alternatively, custom parameters can be provided to define the filter bank.
 
    >>> from maad.features import filter_bank_2d_nodc 
    >>> params, kernels = filter_bank_2d_nodc(frequency=(0.7, 0.5, 0.35, 0.25), ntheta=4, gamma=2, display=True) 
 
    """ 
     
    theta = np.arange(ntheta) 
    theta = theta / ntheta * np.pi 
    params=[i for i in it.product(theta,frequency)] 
    kernels = [] 
    for param in params: 
        kernel = _gabor_kernel_nodc(frequency=param[1], 
                                   theta=param[0], 
                                   bandwidth=bandwidth, 
                                   gamma=gamma, 
                                   offset=0, 
                                   n_stds=3) 
        kernels.append((kernel, param)) 
             
    if display:  
        _, fig = _plot_filter_bank(kernels, frequency, ntheta) 
             
    return params, kernels   
 
#%%
def opt_shape_presets(resolution, opt_shape=None): 
    """  
    Set parameters for multiresolution analysis using presets or custom parameters.
     
    Parameters 
    ---------- 
    resolution: str 
        Select resolution of analysis using presets.  
        Supported presets are: 'low', 'med', and 'high'. 
        Select 'custom' to select user-defined parameters using a dictionary. 
         
    opt_shape: dict 
        Key and values for shape settings. 
        Valid keys are: 'ntheta', 'bandwidth', 'frequency', 'gamma', 'npyr' 
     
    Returns 
    ------- 
    opt_shape: dict 
        A valid dictionary with shape settings 
        
    See also
    --------
    filter_bank_2d_nodc, filter_multires, shape_features
         
    Examples 
    -------- 
    >>> import maad

    Get parameters to analyse at low, med and high shape resolutions.
     
    >>> from maad.features import opt_shape_presets 
    >>> opt_shape_presets('low')  # doctest: +NORMALIZE_WHITESPACE
    {'ntheta': 2,
     'bandwidth': 0.8,
     'frequency': (0.35, 0.5),
     'gamma': 2,
     'npyr': 4}
    
    >>> opt_shape_presets('med')  # doctest: +NORMALIZE_WHITESPACE
    {'ntheta': 4,
     'bandwidth': 0.8,
     'frequency': (0.35, 0.5),
     'gamma': 2,
     'npyr': 6}
    
    >>> opt_shape_presets('high') # doctest: +NORMALIZE_WHITESPACE
    {'ntheta': 8,
     'bandwidth': 0.8,
     'frequency': (0.35, 0.5),
     'gamma': 2,
     'npyr': 6}
    
    """ 
    # Factory presets 
    opt_shape_low = dict(ntheta=2,  
                         bandwidth=0.8,  
                         frequency=(0.35, 0.5),  
                         gamma=2,  
                         npyr = 4) 
    opt_shape_med =  dict(ntheta=4,  
                          bandwidth=0.8,  
                          frequency=(0.35, 0.5),  
                          gamma=2,  
                          npyr = 6) 
    opt_shape_high =  dict(ntheta=8,  
                           bandwidth=0.8,  
                           frequency=(0.35, 0.5),  
                           gamma=2,  
                           npyr = 6) 
     
    if resolution == 'low': 
        opt_shape = opt_shape_low 
     
    elif resolution == 'med': 
       opt_shape = opt_shape_med 
         
    elif resolution == 'high': 
       opt_shape = opt_shape_high 
     
    elif resolution == 'custom': 
        if opt_shape is not None:  # check valid values on opt_shape  
           if all (opt in opt_shape for opt in ('ntheta', 'bandwidth', 'frequency', 'gamma', 'npyr')): 
               pass 
           else: 
               print('Warning: opt_shape must have all keys-values pairs:') 
               print('ntheta, bandwidth, frequency, gamma, npyr') 
               print('Setting resolution to low') 
               opt_shape = opt_shape_low 
 
        else: 
            print('Warning: if resolution is set to custom, a valid opt_shape dictionnary should be provided.') 
            print('Setting resolution to low') 
            opt_shape = opt_shape_low 
    
    else: 
       raise TypeError("Resolution should be: 'low', 'med' or 'high'") 
       opt_shape = opt_shape_low 
 
    return opt_shape 
 
#%%
def shape_features(Sxx, resolution='low', rois=None): 
    """ 
    Computes time-frequency shape coefficients at multiple resolutions using 2D Gabor filters.
     
    Parameters 
    ---------- 
    Sxx: 2D array 
        Input image to process  
    resolution: str or dict 
        Specify resolution of shape descriptors. Can be: 'low', 'med', 'high'. 
        Default is 'low'. Alternatively, custom resolution can be provided  
        using a dictionary with options to define the filter bank. Valid keys  
        are: ntheta, bandwidth, frequency, gamma, npyr.
    rois: pandas DataFrame 
        Regions of interest where descriptors will be computed. Array must  
        have a valid input format with column names: min_t min_f, max_t, max_f. 
        Use format_features(rois,tn,fn) before using shape_features to be sure that 
        the format of the rois DataFrame is correct.
             
    Returns 
    ------- 
    shape: pandas DataFrame 
        Shape coefficient for each region of interest
    params: 2D numpy structured array 
        Corresponding parameters of the 2D fileters used to compute the  
        shape coefficients. Params has 4 fields (theta, freq, pyr_level, scale) 
        
    See Also
    --------
    maad.features.opt_shape_presets, maad.features.filter_bank_2d_nodc, maad.util.plot_shape
    
    References
    ----------
    .. [1] Ulloa, J. S., Aubin, T., Llusia, D., Bouveyron, C., & Sueur, J. (2018). Estimating animal acoustic diversity in tropical environments using unsupervised multiresolution analysis. Ecological Indicators, 90, 346–355. `DOI: 10.1016/j.ecolind.2018.03.026 <https://doi.org/10.1016/j.ecolind.2018.03.026>`_
    .. [2] Sifre, L., & Mallat, S. (2013). Rotation, scaling and deformation invariant scattering for texture discrimination. Computer Vision and Pattern Recognition (CVPR), 2013 IEEE Conference On, 1233–1240. `DOI: 10.1109/CVPR.2013.163 <https://doi.org/10.1109/CVPR.2013.163>`_
    .. [3] Mallat, S. (2008). A Wavelet Tour of Signal Processing: The Sparse Way. Academic Press. `DOI: 10.1016/B978-0-12-374370-1.X0001-8 <https://doi.org/10.1016/B978-0-12-374370-1.X0001-8>`_ 

    Examples 
    -------- 

    Get shape features from the whole power spectrogram 

    >>> from maad.sound import load, spectrogram 
    >>> from maad.features import shape_features
    >>> from maad.util import format_features, power2dB, plot_shape 
    >>> s, fs = load('../data/spinetail.wav')
    >>> Sxx, tn, fn, ext = spectrogram(s, fs, db_range=100, display=True) 
    >>> Sxx_db = power2dB(Sxx, db_range=100)
    >>> shape, params = shape_features(Sxx_db, resolution='med') 
    >>> ax = plot_shape(shape.mean(), params) 
    
    Or get shape features from specific regions of interest 
    
    >>> from maad.util import read_audacity_annot
    >>> rois = read_audacity_annot('../data/spinetail.txt') 
    >>> rois = format_features(rois, tn, fn) 
    >>> shape, params = shape_features(Sxx_db, resolution='med', rois=rois)
    
    The shape coefficient can be visualized using ``plot_shape``. Note that these
    coefficient allow to disciminate between the two vocalization types found in the
    audio file, the spinetail song and the background calls.
    
    >>> shape_crer, shape_sp = shape.groupby('label')
    >>> ax1 = plot_shape(shape_crer[1].filter(regex=r'^shp_').mean(), params) 
    >>> ax1 = ax1.set_title('Spinetail song')
    >>> ax2 = plot_shape(shape_sp[1].filter(regex=r'^shp_').mean(), params) 
    >>> ax2 = ax2.set_title('Background call') # doctest: +NORMALIZE_WHITESPACE

    """     
 
    # Check input data and unpack settings 
    if type(Sxx) is not np.ndarray and len(Sxx.shape) != 2: 
        raise TypeError('Sxx must be a numpy 2D array')   
     
    if type(resolution) is str: 
        opt_shape = opt_shape_presets(resolution) 
    elif type(resolution) is dict: 
        opt_shape = opt_shape_presets(resolution='custom', opt_shape=resolution) 
    else: 
        raise TypeError('Resolution must be string or a dictionary. See function documentation.') 
         
    npyr = opt_shape['npyr']
      
    # build filterbank 
    params, kernels = filter_bank_2d_nodc(ntheta=opt_shape['ntheta'], 
                                          bandwidth=opt_shape['bandwidth'], 
                                          frequency=opt_shape['frequency'], 
                                          gamma=opt_shape['gamma']) 
    # filter spectrogram 
    im_rs = filter_multires(Sxx, kernels, npyr, rescale=True)  
     
    # If rois are provided get mean intensity for each ROI,  
    # else compute mean intensity for the whole spectrogram 
    if rois is None: 
        shape = [] 
        for im in im_rs: 
            shape.append(np.mean(im)) 
        shape = [shape]  # for dataframe formating below
    else: 
        shape = np.zeros(shape=(len(rois),len(im_rs)))
        im_rs = np.array(im_rs)
        index_row = 0 
        for _, row in rois.iterrows(): 
            im_rs_blobs = im_rs[:, int(row.min_y):int(row.max_y)+1, int(row.min_x):int(row.max_x)+1]
            roi_mean = np.mean(im_rs_blobs, axis=(1,2))
            shape[index_row,:] = roi_mean 
            index_row = index_row + 1 
    
    # organise parameters 
    params_multires = _params_to_df(params, npyr) 
     
    # format shape into dataframe 
    cols=['shp_' + str(idx).zfill(3) for idx in range(1,len(shape[0])+1)] 
    shape = pd.DataFrame(data=np.asarray(shape),columns=cols) 
     
    if rois is not None:
        shape.index = rois.index
        shape = rois.join(shape)
         
    return shape, params_multires 
 
#%%
def shape_features_raw(im, resolution='low', opt_shape=None): 
    """ 
    Computes raw shape of 2D signal (image or spectrogram) at multiple resolutions  
    using 2D Gabor filters. Contrary to ``shape_features``, this function 
    delivers the raw response of the spectrogram to each filter of the filter 
    bank. This function can be used to have access to spectrograms filtered by the filterbank.
     
    Parameters 
    ---------- 
    Sxx: 2D array 
        Spectrogram to be analysed 
    resolution:  
        Resolution of analysis, i.e. number of filters used.  
        Three presets are provided, 'low', 'mid' and 'high', which control  
        the number of filters. 
    opt_shape: dictionary (optional) 
        options for the filter bank (kbank_opt) and the number of scales (npyr) 
             
    Returns 
    ------- 
    shape_raw: list 
        Raw shape response of spectrogram to every filter of the filter bank. 
        Every element of the list is the response to one of the filters. Detail  
        of each filter are available in the ``params`` DataFrame returned. 
    params: pandas DataFrame 
        Corresponding parameters of the 2D fileters used to compute the  
        shape coefficient. Params has 4 fields (theta, freq, pyr_level, scale)
    
    See also
    --------
    maad.features.shape_features, maad.features.opt_shape_presets, 
    maad.features.filter_bank_2d_nodc, maad.util.plot_shape
    
    Examples
    --------
    >>> import maad

    Scatter the spectrogram acoustic components using 2D Gabor filters.

    >>> from maad.sound import load, spectrogram 
    >>> from maad.features import shape_features_raw
    >>> from maad.util import power2dB, plot2d
    >>> s, fs = load('../data/spinetail.wav')
    >>> Sxx, tn, fn, ext = spectrogram(s, fs, db_range=80, display=True) 
    >>> Sxx_db = power2dB(Sxx, db_range=80)
    >>> shape_raw, params = shape_features_raw(Sxx_db, resolution='low')
    >>> fig, ax = plot2d(shape_raw[0], **{'extent':ext, 'title': 'High-frequency vertical component'})
    >>> fig, ax = plot2d(shape_raw[13], **{'extent':ext, 'title': 'Low-frequency vertical components'})
    >>> fig, ax = plot2d(shape_raw[2], **{'extent':ext, 'title': 'High-frequency horizontal components'})
    >>> fig, ax = plot2d(shape_raw[15], **{'extent':ext, 'title': 'Low-frequency horizontal components'}) # doctest: +NORMALIZE_WHITESPACE
    """     
     
    # unpack settings 
    opt_shape = opt_shape_presets(resolution, opt_shape) 
    npyr = opt_shape['npyr'] 
    # build filterbank 
    params, kernels = filter_bank_2d_nodc(ntheta=opt_shape['ntheta'], 
                                          bandwidth=opt_shape['bandwidth'], 
                                          frequency=opt_shape['frequency'], 
                                          gamma=opt_shape['gamma']) 
    # filter images 
    shape_raw = filter_multires(im, kernels, npyr, rescale=True)  
     
    # organise parameters 
    params_multires = _params_to_df(params, npyr) 
     
    return shape_raw, params_multires
 
#%%
def centroid_features(Sxx, rois=None, im_rois=None): 
    """ 
    Computes intensity centroid of a spectrogram. If regions of interest 
    ``rois`` are provided, the centroid is computed for each region.
    
    Parameters 
    ---------- 
    Sxx :  2D array 
        Spectrogram 
    rois: pandas DataFrame, default is None 
        Regions of interest where descriptors will be computed. Array must  
        have a valid input format with column names: ``min_t``, ``min_f``, 
        ``max_t``, and ``max_f``. Use the function ``maad.util.format_features``
        before using centroid_features to format of the ``rois`` DataFrame 
        correctly.
    im_rois: 2d ndarray 
        image with labels as values
             
    Returns 
    ------- 
    centroid: pandas DataFrame 
        Centroid of each region of interest.
        
    See Also
    --------
    maad.features.shape_features, maad.util.overlay_rois, 
    maad.util.format_features
 
    Examples
    --------
    >>> import maad
 
    Get centroid from the whole power spectrogram 
 
    >>> from maad.sound import load, spectrogram
    >>> from maad.features import centroid_features
    >>> from maad.util import (power2dB, format_features, overlay_rois, plot2d, overlay_centroid)
     
    Load audio and compute spectrogram 
     
    >>> s, fs = load('../data/spinetail.wav') 
    >>> Sxx,tn,fn,ext = spectrogram(s, fs, db_range=80) 
    >>> Sxx = power2dB(Sxx, db_range=80)
     
    Load annotations and plot
    
    >>> from maad.util import read_audacity_annot
    >>> rois = read_audacity_annot('../data/spinetail.txt') 
    >>> rois = format_features(rois, tn, fn) 
    >>> ax, fig = plot2d (Sxx, extent=ext)
    >>> ax, fig = overlay_rois(Sxx,rois, extent=ext, ax=ax, fig=fig)
    
    Compute the centroid of each rois, format to get results in the 
    temporal and spectral domain and overlay the centroids.
     
    >>> centroid = centroid_features(Sxx, rois) 
    >>> centroid = format_features(centroid, tn, fn)
    >>> ax, fig = overlay_centroid(Sxx,centroid, extent=ext, ax=ax, fig=fig)
    
    """     
     
    # Check input data 
    if type(Sxx) is not np.ndarray and len(Sxx.shape) != 2: 
        raise TypeError('Sxx must be an numpy 2D array')       
     
    # check rois 
    if rois is not None: 
        if not(('min_t' and 'min_f' and 'max_t' and 'max_f') in rois): 
            raise TypeError('Array must be a Pandas DataFrame with column names: min_t, min_f, max_t, max_f. Check example in documentation.') 
     
    centroid=[] 
    area = []   
    if rois is None: 
        centroid = ndimage.center_of_mass(Sxx) 
        centroid = pd.DataFrame(np.asarray(centroid)).T 
        centroid.columns = ['centroid_y', 'centroid_x'] 
        centroid['area_xy'] = Sxx.shape[0] * Sxx.shape[1]
        centroid['duration_x'] = Sxx.shape[1]
        centroid['bandwidth_y'] = Sxx.shape[0]
    else: 
        if im_rois is not None : 
            # real centroid and area
            rprops = measure.regionprops(im_rois, intensity_image=Sxx)
            centroid = [roi.weighted_centroid for roi in rprops]
            area = [roi.area for roi in rprops]
        else:
            # rectangular area (overestimation) 
            area = (rois.max_y -rois.min_y) * (rois.max_x -rois.min_x)  
            # centroid of rectangular roi
            for _, row in rois.iterrows() : 
                row = pd.DataFrame(row).T 
                im_blobs = rois_to_imblobs(np.zeros(Sxx.shape), row)     
                rprops = measure.regionprops(im_blobs, intensity_image=Sxx)
                centroid.append(rprops.pop().weighted_centroid) 

        centroid = pd.DataFrame(centroid, columns=['centroid_y', 'centroid_x'], index=rois.index)
        
        ##### duration in number of pixels 
        centroid['duration_x'] = (rois.max_x -rois.min_x)  
        ##### bandwidth in number of pixels 
        centroid['bandwidth_y'] = (rois.max_y -rois.min_y) 
        ##### area
        centroid['area_xy'] = area      
     
        # concat rois and centroid dataframes 
        centroid = rois.join(pd.DataFrame(centroid, index=rois.index))  

    return centroid 
 
#%%     
def all_shape_features(s, fs, rois, resolution='low',  
                          display=False, **kwargs): 
    """ 
    Computes shape and central frequency features from signal at specified 
    time-frequency limits defined by regions of interest (ROIs).
     
    Parameters 
    ---------- 
    s: ndarray 
        Singal to be analysed 
    fs: int 
        Sampling frequency of the signal 
    rois: pandas DataFrame 
        Time frequency limits for the analysis. Columns should have at 
        least min_t, max_t, min_f, max_f. Can be computed with multiple 
        detection methods, such as find_rois_cwt 
    resolution: str or dict, default is 'med' 
        Specify resolution of shape descriptors. Can be: 'low', 'med', 'high'. 
        Default is 'low'. Alternatively, custom resolution can be provided  
        using a dictionary with options to define the filter bank. Valid keys  
        are: ntheta, bandwidth, frequency, gamma, npyr         
    opt_spec: dictionnary, optional 
        Options for the spectrogram with keys, window length 'nperseg', 
        window overlap in percentage 'overlap' and frequency limits 'flims' 
        which is a list of 2 scalars : minimum and maximum boundary frequency  
        values in Hertz 
    display: boolean, optional, default is False 
        Flag. If display is True, plot results 
        
    kwargs, optional. This parameter is used by plt.plot and savefig functions 
            
        - savefilename : str, optional, default :'_spectro_overlaycentroid.png' 
            Postfix of the figure filename 
         
        - figsize : tuple of integers, optional, default: (4,10) 
            width, height in inches.   
         
        - title : string, optional, default : 'Spectrogram' 
            title of the figure 
         
        - xlabel : string, optional, default : 'Time [s]' 
            label of the horizontal axis 
         
        - ylabel : string, optional, default : 'Amplitude [AU]' 
            label of the vertical axis 
         
        - cmap : string or Colormap object, optional, default is 'gray' 
            See https://matplotlib.org/examples/color/colormaps_reference.html 
            in order to get all the  existing colormaps 
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic',  
            'viridis'... 
         
        - vmin, vmax : scalar, optional, default: None 
            `vmin` and `vmax` are used in conjunction with norm to normalize 
            luminance data.  Note if you pass a `norm` instance, your 
            settings for `vmin` and `vmax` will be ignored. 
                 
        - dpi : integer, optional, default is 96 
            Dot per inch.  
            For printed version, choose high dpi (i.e. dpi=300) => slow 
            For screen version, choose low dpi (i.e. dpi=96) => fast 
         
        - format : string, optional, default is 'png' 
            Format to save the figure  
         
        ... and more, see matplotlib
        
    Returns 
    ------- 
    feature_rois: pandas Dataframe 
        A dataframe with each column corresponding to a feature 
     
    Examples 
    -------- 
    >>> import maad
    >>> from maad.util import read_audacity_annot 
    >>> from maad.sound import load 
    >>> from maad.features import all_shape_features 
    >>> s, fs = load('../data/spinetail.wav')
    >>> rois = read_audacity_annot('../data/spinetail.txt')
    >>> features = all_shape_features(s, fs, rois, resolution='low', display=True)  # doctest: +NORMALIZE_WHITESPACE
        number of rois : 18
    >>> print(features.iloc[0]) # doctest: +NORMALIZE_WHITESPACE
        label                    SP
        min_t              0.101385
        min_f           6441.064453
        max_t               0.36752
        max_f          12296.577148
        min_y                   150
        min_x                     8
        max_y                   286
        max_x                    31
        shp_001            0.160656
        shp_002            0.155573
        shp_003            0.064787
        shp_004            0.066215
        shp_005            0.158675
        shp_006            0.125203
        shp_007            0.044987
        shp_008            0.043188
        shp_009            0.218268
        shp_010            0.158725
        shp_011            0.043553
        shp_012            0.031206
        shp_013            0.268222
        shp_014             0.21718
        shp_015            0.059112
        shp_016            0.038394
        centroid_y            220.0
        centroid_x             19.0
        duration_x       135.999274
        bandwidth_y             0.0
        area_xy               12512
        centroid_f      9474.609375
        centroid_t           0.2322
        duration_t         0.267029
        bandwidth_f      5857.03125
        area_tf              1564.0
        Name: 0, dtype: object
    >>> features.columns # doctest: +NORMALIZE_WHITESPACE
        Index(['label', 'min_t', 'min_f', 'max_t', 'max_f', 'min_y', 'min_x', 'max_y',
            'max_x', 'shp_001', 'shp_002', 'shp_003', 'shp_004', 'shp_005',
            'shp_006', 'shp_007', 'shp_008', 'shp_009', 'shp_010', 'shp_011',
            'shp_012', 'shp_013', 'shp_014', 'shp_015', 'shp_016', 'centroid_y',
            'centroid_x', 'duration_x', 'bandwidth_y', 'area_xy', 'centroid_f',
            'centroid_t', 'duration_t', 'bandwidth_f', 'area_tf'],
            dtype='object')

    """    
   
    # Spectro computing
    N=1024
    window  = kwargs.pop('window','hann')
    nperseg = kwargs.pop('nperseg',N) 
    overlap = kwargs.pop('overlap',N//2)  
    mode    = kwargs.pop('mode', 'psd')   
    fcrop   = kwargs.pop('fcrop',None) 
    tcrop   = kwargs.pop('tcrop',None)
    verbose = kwargs.pop('verbose',False) 
    # Spectro display
    ax      = kwargs.pop('ax',None)  
    fig     = kwargs.pop('fig',None)  
    vmin   = kwargs.pop('vmin',-120) 
    vmax   = kwargs.pop('vmax',-20)
    
    Sxx, tn, fn, ext = spectrogram(s, fs, window, nperseg, overlap, fcrop,
                                   tcrop, mode, verbose, display=False, **kwargs) 
    
    # dB scale
    Sxx = 10*np.log10(Sxx)
     
    # format rois to bbox 
    rois = format_features(rois, tn, fn) 
    
    # if rois becomes empty after format_features
    if rois.empty :
        print ('rois is empty')
        features = rois
        return features
    else:
        print('number of rois : %d' % len(rois))
     
    # Compute shape features and centroid features 
    shape, params = shape_features(Sxx, resolution, rois) 
    shape = format_features(shape, tn, fn)
    if verbose : 
        print('Dataframe with shapes features')
        print(shape)
        
    centroid = centroid_features(Sxx, rois) 
    centroid = format_features(centroid, tn, fn)
    if verbose : 
        print('Dataframe with centroids features')
        print(centroid)
    
    if display :
        # view bbox
        ax, fig = overlay_rois(Sxx, rois, vmin=vmin, vmax=vmax, **kwargs)
        # view centroids
        overlay_centroid(Sxx, centroid, savefig=None, 
                         fig=fig, ax=ax, **kwargs)
        # plot shape
        plot_shape(shape.filter(regex=r'^shp_').mean(), params)
    
    # combine into a single df without columns duplication
    features = pd.concat([rois, shape, centroid], axis=1, sort=False)
    features = features.loc[:,~features.columns.duplicated()]
    
    if verbose : 
        print('list of features')
        print(list(features))
     
    return features 

if __name__ == "__main__":
    import doctest
    doctest.testmod()


#!/usr/bin/env python
""" 
Collection of metrics functions to evaluate audio preprocessing methods
"""
#
# Authors:  Juan Sebastian ULLOA <lisofomia@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

# =============================================================================
# Load the modules
# =============================================================================
# Import external modules
import numpy as np

# Import external modules
from maad.sound import envelope, avg_power_spectro, remove_background_along_axis
from maad.util import power2dB, mean_dB, get_unimode, running_mean

#%%
# =============================================================================
# public functions
# =============================================================================
def temporal_snr (s, mode ='fast', Nt=512) :
    """
    Compute the signal to noise ratio (SNR) of an audio signal in the time domain.

    Parameters
    ----------
    s : 1D array
        Audio to process
    mode : str, optional, default is `fast`
        Select the mode to compute the envelope of the audio waveform
        `fast` : The sound is first divided into frames (2d) using the 
        function _wave2timeframes(s), then the max of each frame gives a 
        good approximation of the envelope.
        `Hilbert` : estimation of the envelope from the Hilbert transform. 
        The method is slow
    Nt : integer, optional, default is 512
        Size of each frame. The largest, the highest is the approximation.
    
    Returns
    -------
    ENRt : float
        Total energy in dB computed in the time domain
    BGNt : float
        Estimation of the background energy (dB) computed in the time domain
    SNRt: float
        Signal to noise ratio (dB) computed in the time domain 
        SNRt = ENRt - BGNt

    References
    ----------
    .. [1] Towsey, Michael (2013), Noise Removal from Waveforms and Spectrograms Derived from Natural Recordings of the Environment. Queensland University of Technology, Brisbane.
    .. [2] Towsey, Michael (2017),The calculation of acoustic indices derived from long-duration recordings of the naturalenvironment. Queensland University of Technology, Brisbane.

    Examples
    --------
    >>> s, fs = maad.sound.load('../data/rock_savanna.wav')
    >>> _,_,snr = maad.sound.temporal_snr(s)
    >>> snr
    1.5744987447774665

    """
    # Envelope
    env = envelope(s, mode=mode, Nt=Nt)
    # linear to power dB
    envdB = power2dB(env**2)
    # total energy estimation. 
    ENRt = mean_dB(envdB, axis=1)
    # Background noise estimation
    BGNt = get_unimode (envdB, mode ='median')
    # Signal to Noise ratio estimation
    SNRt = ENRt - BGNt

    return ENRt, BGNt, SNRt

#%%
def spectral_snr (Sxx_power) :
    """
    Compute the signal to noise ratio (SNR) of an audio from its spectrogram
    in the time-frequency domain.

    Parameters
    ----------
    Sxx_power : 2D array
        Power spectrogram density [Matrix] to process.
        
    Returns
    -------
    ENRf : float
        Total energy in dB computed in the frequency domain which corresponds
        to the average of the power spectrogram then the sum of the average
    BGNf : float
        Estimation of the background energy (dB) computed based on the estimation
        of the noise profile of the power spectrogram (2d)
    SNRf: float
        Signal to noise ratio (dB) computed in the frequency domain 
        SNRf = ENRf - BGNf
    ENRf_per_bin : vector of floats
        Energy in dB per frequency bin
    BGNf_per_bin : vector of floats
        Background (noise profile) energy in dB per frequency bin
    SNRf_per_bin : vector of floats  
        Signal to noise ratio per frequency bin
        
    References
    ----------
    ..[1] Towsey, Michael (2013), Noise Removal from Waveforms and Spectrograms Derived from Natural Recordings of the Environment. Queensland University of Technology, Brisbane.
    ..[2] Towsey, Michael (2017),The calculation of acoustic indices derived from long-duration recordings of the naturalenvironment.Queensland University of Technology, Brisbane.

    Examples
    --------
    >>> s, fs = maad.sound.load('../data/rock_savanna.wav')
    >>> Sxx_power,_,_,_ = maad.sound.spectrogram (s, fs)  
    >>> _, _, snr, _, _, _ = maad.sound.spectral_snr(Sxx_power)
    >>> snr
    4.084065436435541
    
    """
    # average Sxx_power along time axis
    ENRf_per_bin = avg_power_spectro(Sxx_power)
    # compute total energy in dB
    ENRf = power2dB(sum(ENRf_per_bin))
    # Extract the noise profile (BGNf_per_bin) from the spectrogram Sxx_power
    _, noise_profile = remove_background_along_axis(Sxx_power, mode='median',axis=1) 
    # smooth the profile by removing spurious thin peaks (less than 5 pixels wide)
    noise_profile = running_mean(noise_profile,N=5)
    # noise_profile (energy) into dB
    BGNf_per_bin= power2dB(noise_profile)
    # compute noise/background energy in dB
    BGNf = power2dB(sum(noise_profile))
    # compute signal to noise ratio
    SNRf = ENRf - BGNf 
    # compute SNR_per_bin
    SNRf_per_bin = ENRf_per_bin - ENRf_per_bin

    return ENRf, BGNf, SNRf, ENRf_per_bin, BGNf_per_bin,SNRf_per_bin 

#%%
def sharpness (Sxx) :
    """ 
    Compute the sharpness of a spectrogram
     
    Parameters 
    ---------- 
    Sxx : 2d ndarray of scalars 
        Spectrogram (or image) 
        
    Returns
    -------
    sharpness : scalar
        sharpness of the spectrogram (or image)
        
    Examples
    --------
    >>> s, fs = maad.sound.load('../data/rock_savanna.wav')
    >>> Sxx_power,_,_,_ = maad.sound.spectrogram (s, fs)  
    >>> sharp = maad.sound.sharpness(Sxx_power)
    >>> sharp
    1.1930709869950632e-05
    
    """
    
    Gt = np.gradient(Sxx, edge_order=1, axis=1)
    Gf = np.gradient(Sxx, edge_order=1, axis=0)
    S = np.sqrt(Gt**2+ Gf**2)
    sharpness=sum(sum(S))/(Gt.shape[0]*Gt.shape[1])
    return sharpness   
 



# -*- coding: utf-8 -*-
""" 
Acoustic features
=================

The module ``features`` is an ensemble of functions to characterize audio signals using temporal and spectral features, and ecoacoustic indices.


Spectro-temporal features
-------------------------
.. autosummary::
    :toctree: generated/
    
    shape_features
    shape_features_raw
    opt_shape_presets
    filter_multires
    filter_bank_2d_nodc
    centroid_features
    all_shape_features

Alpha acoustic indices
----------------------
.. autosummary::
    :toctree: generated/
        
    temporal_median
    temporal_entropy
    acoustic_richness_index
    temporal_activity
    temporal_events
    acoustic_complexity_index
    frequency_entropy
    number_of_peaks
    spectral_entropy
    spectral_activity
    spectral_events
    spectral_cover
    soundscape_index
    bioacoustics_index
    acoustic_diversity_index
    acoustic_eveness_index
    roughness
    temporal_leq
    spectral_leq
    surface_roughness
    tfsd
    more_entropy
    acoustic_gradient_index
    frequency_raoq
    region_of_interest_index_deprecated
    region_of_interest_index
    all_temporal_alpha_indices
    all_spectral_alpha_indices

Temporal features
-----------------
.. autosummary::
    :toctree: generated/
    
    temporal_moments
    zero_crossing_rate
    temporal_duration
    temporal_quantile
    all_temporal_features
    
Spectral features
-----------------
.. autosummary::
    :toctree: generated/
        
    spectral_moments
    peak_frequency
    spectral_quantile
    spectral_bandwidth
    all_spectral_features

Composite acoustic features
---------------------------
.. autosummary::
    :toctree: generated/

    graphical_soundscape
    plot_graph    

"""

from .shape import (filter_multires,
                    filter_bank_2d_nodc,
                    opt_shape_presets,
                    shape_features,
                    shape_features_raw,
                    centroid_features,
                    all_shape_features)

from .spectral import (spectral_moments,
                    peak_frequency,
                    spectral_quantile,
                    spectral_bandwidth,
                    all_spectral_features)

from .temporal import (temporal_moments,
                    zero_crossing_rate,
                    temporal_duration,
                    temporal_quantile,
                    all_temporal_features)

from .alpha_indices import (temporal_median,
                            temporal_entropy,
                            acoustic_richness_index,
                            temporal_activity,
                            temporal_events,
                            acoustic_complexity_index,
                            frequency_entropy,
                            number_of_peaks,
                            spectral_entropy,
                            spectral_activity,
                            spectral_events,
                            spectral_cover,
                            soundscape_index,
                            bioacoustics_index,
                            acoustic_diversity_index,
                            acoustic_eveness_index,
                            roughness,
                            temporal_leq,
                            spectral_leq,
                            surface_roughness,
                            tfsd,
                            more_entropy,
                            acoustic_gradient_index,
                            frequency_raoq,
                            region_of_interest_index_deprecated,
                            region_of_interest_index,
                            all_temporal_alpha_indices,
                            all_spectral_alpha_indices)

from .composite_soundscape_descriptors import (graphical_soundscape,
                                            plot_graph)

__all__ = [
        # shape
        'filter_multires', 
        'filter_bank_2d_nodc',
        'opt_shape_presets',
        'shape_features',
        'shape_features_raw',
        'centroid_features',
        'all_shape_features',
        
        # spectral
        'spectral_moments',
        'peak_frequency',
        'spectral_quantile',
        'temporal_quantile',
        'spectral_bandwidth',
        'all_spectral_features',
        
        # temporal
        'temporal_moments',
        'zero_crossing_rate',
        'temporal_duration',
        'all_temporal_features',
        
        # alpha_indices
        'temporal_moments',
        'temporal_median',
        'temporal_entropy',
        'acoustic_richness_index',
        "temporal_activity",
        "temporal_events",
        "acoustic_complexity_index",
        "frequency_entropy",
        "number_of_peaks",
        "spectral_entropy",
        "spectral_moments",
        "spectral_activity",
        "spectral_events",
        'spectral_cover',
        "soundscape_index",
        'bioacoustics_index',
        "acoustic_diversity_index",
        "acoustic_eveness_index",
        "roughness",
        'temporal_leq',
        'spectral_leq',
        "surface_roughness",
        "tfsd",
        "more_entropy",
        "acoustic_gradient_index",
        "frequency_raoq",
        "region_of_interest_index_deprecated",
        "region_of_interest_index",
        'all_temporal_alpha_indices',
        'all_spectral_alpha_indices',
        
        # composite features
        'graphical_soundscape',
        'plot_graph',
        ]


#!/usr/bin/env python
"""
Collection of functions to remove background noise from spectrogram using
spectral subtraction methods
"""
#
# Authors:  Juan Sebastian ULLOA <lisofomia@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

# =============================================================================
# Load the modules
# =============================================================================
# Import external modules
from maad.util import (plot1d, plot2d, running_mean,
                       get_unimode, mean_dB, power2dB)
import matplotlib.pyplot as plt
import numpy as np
from scipy import ndimage
from skimage.morphology import reconstruction
from scipy import signal
# min value
import sys
_MIN_ = sys.float_info.min

# Import internal modules

# %%
# =============================================================================
# public functions
# =============================================================================

def remove_background(Sxx, gauss_win=50, gauss_std=25, beta1=1, beta2=1,
                      llambda=1, verbose=False, display = False,
                      savefig=None, **kwargs):
    """ 
    Remove background noise using spectral subtraction.

    Based on the spectrum of the A posteriori noise profile.   
    It computes an atenuation map in the time-frequency domain.  
    See [1]_ or [2]_ for more detail about the algorithm. 

    Parameters 
    ---------- 
    Sxx : 2d ndarray of scalars 
        Spectrogram  

    gauss_win=50 : int, optional, default: 50 
        Number of points in the gaussian window  

    gauss_std = 25 
        The standard deviation, sigma used to create the gaussian window   

    beta1 : scaler, optional, default: 1       
        beta1 has to be >0 
        Should be close to 1 

    beta2: scaler, optional, default: 1        
        beta2 has to be >0  
        better to not change 

    llambda : int, optional, default: 1        
        over-subtraction factor to compensate variation of noise amplitude. 
        Should be close to 1 

    verbose : boolean, optional, default is False
        Print messages and speed

    display : boolean, optional, default is False 
        Display the signal if True 

    savefig : string, optional, default is None 
        Root filename (with full path) is required to save the figures. Postfix 
        is added to the root filename. 

    kwargs, optional. This parameter is used by plt.plot and savefig functions 

        - savefilename : str, optional, default :'_spectro_after_noise_subtraction.png' 
            Postfix of the figure filename 

        - figsize : tuple of integers, optional, default: (4,10) 
            width, height in inches.   

        - title : string, optional, default : 'Spectrogram' 
            title of the figure 

        - xlabel : string, optional, default : 'Time [s]' 
            label of the horizontal axis 

        - ylabel : string, optional, default : 'Amplitude [AU]' 
            label of the vertical axis 

        - cmap : string or Colormap object, optional, default is 'gray' 
            See https://matplotlib.org/examples/color/colormaps_reference.html 
            in order to get all the  existing colormaps 
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic',  
            'viridis'... 

        - vmin, vmax : scalar, optional, default: None 
            `vmin` and `vmax` are used in conjunction with norm to normalize 
            luminance data.  Note if you pass a `norm` instance, your 
            settings for `vmin` and `vmax` will be ignored. 

        - extent : scalars (left, right, bottom, top), optional, default: None 
            The location, in data-coordinates, of the lower-left and 
            upper-right corners. If `None`, the image is positioned such that 
            the pixel centers fall on zero-based (row, column) indices. 

        - dpi : integer, optional, default is 96 
            Dot per inch.  
            For printed version, choose high dpi (i.e. dpi=300) => slow 
            For screen version, choose low dpi (i.e. dpi=96) => fast 

        - format : string, optional, default is 'png' 
            Format to save the figure 

        ... and more, see matplotlib     

    Returns 
    ------- 
    Sxx_out : 2d ndarray of scalar 
        Spectrogram after denoising   
    noise_profile : 1d darray of scalar
        noise_profile
    BGNxx : 2d ndarray of scalar 
        Noise map

    References 
    ---------- 
    .. [1] Steven F. Boll, "Suppression of Acoustic Noise in Speech Using Spectral 
       Subtraction", IEEE Transactions on Signal Processing, 27(2),pp 113-120,1979  
       `DOI:10.1109/TASSP.1979.1163209 <https://doi.org/10.1109/TASSP.1979.1163209>`_
       
    .. [2] Y. Ephraim and D. Malah, Speech enhancement using a minimum mean square  
       error short-time spectral amplitude estimator, IEEE. Transactions in 
       Acoust., Speech, Signal Process., vol. 32, no. 6, pp. 11091121, Dec. 1984.   
       `DOI:10.1109/TASSP.1984.1164453 <https://doi.org/10.1109/TASSP.1984.1164453>`_
       
    Examples
    --------

    Load audio recording and convert it into spectrogram

    >>> s, fs = maad.sound.load('../data/rock_savanna.wav')
    >>> Sxx,tn,fn,ext = maad.sound.spectrogram (s, fs)   

    Convert linear spectrogram into dB and add 96dB (which is the maximum dB
    for 16 bits wav) in order to have positive values

    >>> Sxx_dB = maad.util.power2dB(Sxx) + 96

    Remove stationnary noise from the spectrogram in dB

    >>> Sxx_dB_noNoise, noise_profile, _ = maad.sound.remove_background(Sxx_dB)

    Plot both spectrograms

    >>> import matplotlib.pyplot as plt 
    >>> import numpy as np
    >>> fig, (ax1, ax2) = plt.subplots(2, 1)
    >>> maad.util.plot2d(Sxx_dB, ax=ax1, extent=ext, title='original', vmin=np.median(Sxx_dB), vmax=np.median(Sxx_dB)+40)
    >>> maad.util.plot2d(Sxx_dB_noNoise, ax=ax2, extent=ext, title='Without stationary noise', vmin=np.median(Sxx_dB_noNoise), vmax=np.median(Sxx_dB_noNoise)+40)
    >>> fig.set_size_inches(15,8)
    >>> fig.tight_layout()

    """
    if verbose:
        print(72 * '_')
        print('Determine the profile of the stochastic background noise...')

    Nf, Nw = Sxx.shape

    # average spectrum (assumed to be ergodic)
    mean_profile = np.mean(Sxx, 1)

    # White Top Hat (to remove non uniform background) = i - opening(i)
    selem = signal.windows.gaussian(gauss_win, gauss_std)
    noise_profile = ndimage.grey_opening(mean_profile, structure=selem)

    # Remove the artefact at the end of the spectrum (2 highest frequencies)
    noise_profile[-2:] = mean_profile[-2:]
    noise_profile[:2] = mean_profile[:2]

    # Create a matrix with the noise profile
    noise_spectro = np.kron(np.ones((Nw, 1)), noise_profile)
    noise_spectro = noise_spectro.transpose()

    # snr estimate a posteriori
    SNR_est = Sxx - noise_spectro
    # to avoid dividing by 0
    SNR_est[SNR_est <= 0] = 0
    noise_spectro[noise_spectro == 0] = _MIN_
    # ratio
    SNR_est = (Sxx/noise_spectro)
    # keep only positive values
    SNR_est = SNR_est*(SNR_est > 0)

    # compute attenuation map
    # if llambda, beta1 and beta 2 are equal to 1, it is (1 - noise_spectro)
    an_lk = (1-llambda*((1./(SNR_est+1))**beta1))**beta2
    an_lk = an_lk*(an_lk > 0)  # keep only positive values

    if verbose:
        print('Remove the stochastic background noise...')

    # Apply the attenuation map to the STFT coefficients
    Sxx_out = an_lk*Sxx

    # noise map BGNxx
    BGNxx = Sxx - Sxx_out

    # if nan in the image, convert nan into 0
    np.nan_to_num(Sxx_out, 0)

    # Set negative value to 0
    Sxx_out[Sxx_out < 0] = 0

    # Display
    if display:
        ylabel = kwargs.pop('ylabel', 'Frequency [Hz]')
        xlabel = kwargs.pop('xlabel', 'Time [sec]')
        title = kwargs.pop('title', 'Spectrogram without stationnary noise')
        cmap = kwargs.pop('cmap', 'gray')
        vmin = kwargs.pop('vmin', np.min(Sxx_out))
        vmax = kwargs.pop('vmax', np.max(Sxx_out))
        extent = kwargs.pop('extent', None)

        if extent is not None:
            fn = np.arange(0, Nf)*(extent[3]-extent[2])/(Nf-1) + extent[2]
            xlabel = 'frequency [Hz]'
            figsize = kwargs.pop('figsize', (4, 0.33*(extent[1]-extent[0])))
        else:
            fn = np.arange(Nf)
            xlabel = 'pseudofrequency [points]'
            figsize = kwargs.pop('figsize', (4, 13))

        _, fig = plot2d(Sxx_out, extent=extent, figsize=figsize, title=title,
                        ylabel=ylabel, xlabel = xlabel, vmin=vmin, vmax=vmax,
                        cmap=cmap, **kwargs)

        fig2, (ax1, ax2) = plt.subplots(2, sharex=True)
        fig2.set_size_inches((5, 4))
        ax1, _ = plot1d(fn, mean_profile, ax=ax1, legend='Original profile',
                        color='b',
                        xlabel='', ylabel= 'Amplitude [dB]', figtitle='')
        ax1, _ = plot1d(fn, np.mean(BGNxx, axis=1), ax=ax1, legend='Noise profile',
                        color='r',
                        xlabel='', ylabel='Amplitude [dB]', figtitle='')
        ax2, _ = plot1d(fn, np.mean(Sxx_out, axis=1), ax=ax2, color='k',
                        legend='Denoized profile',
                        xlabel=xlabel, ylabel= 'Amplitude [dB]', figtitle='')
        fig2.tight_layout()

        # SAVE FIGURE
        if savefig is not None:
            dpi = kwargs.pop('dpi', 96)
            dpi = kwargs.pop('dpi', 96)
            bbox_inches = kwargs.pop('bbox_inches', 'tight')
            format = kwargs.pop('format', 'png')
            savefilename = kwargs.pop(
                'savefilename', '_spectro_after_noise_subtraction')
            filename = savefig+savefilename+'.'+format
            print('\n''save figure : %s' % filename)
            fig.savefig(fname=filename, dpi=dpi, bbox_inches=bbox_inches,
                        format=format, **kwargs)

    return Sxx_out, noise_profile, BGNxx

# %%


def remove_background_morpho(Sxx, q=0.1, display=False, savefig=None, **kwargs):
    """ 
    Remove background noise in a spectrogram using mathematical morphology tool.

    Parameters 
    ---------- 
    Sxx : 2D numpy array  
        Original spectrogram (or image) 

    q : float
        Quantile which must be between  0 and 1 inclusive. The closest to one, 
        the finest details are kept

    display : boolean, optional, default is False 
        Display the signal if True 

    savefig : string, optional, default is None 
        Root filename (with full path) is required to save the figures. Postfix 
        is added to the root filename. 

    kwargs, optional. This parameter is used by plt.plot and savefig functions 

        - savefilename : str, optional, default :'_spectro_after_noise_subtraction.png' 
            Postfix of the figure filename 

        - figsize : tuple of integers, optional, default: (4,10) 
            width, height in inches.   

        - title : string, optional, default : 'Spectrogram' 
            title of the figure 

        - xlabel : string, optional, default : 'Time [s]' 
            label of the horizontal axis 

        - ylabel : string, optional, default : 'Amplitude [AU]' 
            label of the vertical axis 

        - cmap : string or Colormap object, optional, default is 'gray' 
            See https://matplotlib.org/examples/color/colormaps_reference.html 
            in order to get all the  existing colormaps 
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic',  
            'viridis'... 

        - vmin, vmax : scalar, optional, default: None 
            `vmin` and `vmax` are used in conjunction with norm to normalize 
            luminance data.  Note if you pass a `norm` instance, your 
            settings for `vmin` and `vmax` will be ignored. 

        - extent : scalars (left, right, bottom, top), optional, default: None 
            The location, in data-coordinates, of the lower-left and 
            upper-right corners. If `None`, the image is positioned such that 
            the pixel centers fall on zero-based (row, column) indices. 

        - dpi : integer, optional, default is 96 
            Dot per inch.  
            For printed version, choose high dpi (i.e. dpi=300) => slow 
            For screen version, choose low dpi (i.e. dpi=96) => fast 

        - format : string, optional, default is 'png' 
            Format to save the figure 

        ... and more, see matplotlib     

    Returns 
    ------- 
    Sxx_out : 2d ndarray of scalar 
        Spectrogram after denoising   
    noise_profile : 1d ndarray of scalar
        Noise profile
    BGNxx : 2d ndarray of scalar 
        Noise map

    Examples
    --------

    Load audio recording and convert it into spectrogram

    >>> s, fs = maad.sound.load('../data/rock_savanna.wav')
    >>> Sxx,tn,fn,ext = maad.sound.spectrogram (s, fs)   

    Convert linear spectrogram into dB

    >>> Sxx_dB = maad.util.power2dB(Sxx) +96

    Remove stationnary noise from the spectrogram 

    >>> Sxx_dB_noNoise,_,_ = maad.sound.remove_background_morpho(Sxx_dB, q=0.5)

    Plot both spectrograms

    >>> import matplotlib.pyplot as plt 
    >>> import numpy as np
    >>> fig, (ax1, ax2) = plt.subplots(2, 1)
    >>> maad.util.plot2d(Sxx_dB, ax=ax1, extent=ext, title='original', vmin=np.median(Sxx_dB), vmax=np.median(Sxx_dB)+40)
    >>> maad.util.plot2d(Sxx_dB_noNoise, ax=ax2, extent=ext, title='Without stationary noise',vmin=np.median(Sxx_dB_noNoise), vmax=np.median(Sxx_dB_noNoise)+40)
    >>> fig.set_size_inches(15,8)
    >>> fig.tight_layout()     

    Load audio recording and convert it into spectrogram

    >>> s, fs = maad.sound.load('../data/rock_savanna.wav')
    >>> Sxx,tn,fn,ext = maad.sound.spectrogram (s, fs, tcrop=(0,20))   
    >>> Sxx_dB = maad.util.power2dB(Sxx) +96

    Remove stationnary noise from the spectrogram with different q

    >>> Sxx_dB_noNoise_q25,_,_ = maad.sound.remove_background_morpho(Sxx_dB, q=0.25)
    >>> Sxx_dB_noNoise_q50,_,_ = maad.sound.remove_background_morpho(Sxx_dB, q=0.5)
    >>> Sxx_dB_noNoise_q75,_,_ = maad.sound.remove_background_morpho(Sxx_dB, q=0.75)

    Plot 3 spectrograms

    >>> import matplotlib.pyplot as plt 
    >>> fig, (ax1, ax2, ax3) = plt.subplots(3, 1)
    >>> maad.util.plot2d(Sxx_dB_noNoise_q25, ax=ax1, extent=ext, title='Without stationary noise (q=0.25)',vmin=np.median(Sxx_dB_noNoise_q25), vmax=np.median(Sxx_dB_noNoise_q25)+40)
    >>> maad.util.plot2d(Sxx_dB_noNoise_q50, ax=ax2, extent=ext, title='Without stationary noise (q=0.50)',vmin=np.median(Sxx_dB_noNoise_q50), vmax=np.median(Sxx_dB_noNoise_q50)+40)
    >>> maad.util.plot2d(Sxx_dB_noNoise_q75, ax=ax3, extent=ext, title='Without stationary noise (q=0.75)',vmin=np.median(Sxx_dB_noNoise_q75), vmax=np.median(Sxx_dB_noNoise_q75)+40)
    >>> fig.set_size_inches(15,9)
    >>> fig.tight_layout()     

    """

    # Use morpho math tools to estimate the background noise
    BGNxx = reconstruction(seed=Sxx-(np.quantile(Sxx, q)),
                           mask=Sxx, method='dilation')
    Sxx_out = Sxx - BGNxx

    # noise profile along time axis
    noise_profile = np.mean(BGNxx, 1)

    # Set negative value to 0
    Sxx_out[Sxx_out < 0] = 0

    # Display
    if display:
        ylabel = kwargs.pop('ylabel', 'Frequency [Hz]')
        xlabel = kwargs.pop('xlabel', 'Time [sec]')
        title = kwargs.pop('title', 'Spectrogram without stationnary noise')
        cmap = kwargs.pop('cmap', 'gray')
        vmin = kwargs.pop('vmin', np.min(Sxx_out))
        vmax = kwargs.pop('vmax', np.max(Sxx_out))
        extent = kwargs.pop('extent', None)

        Nf, Nw = Sxx.shape

        if extent is not None:
            fn = np.arange(0, Nf)*(extent[3]-extent[2])/(Nf-1) + extent[2]
            xlabel = 'frequency [Hz]'
            figsize = kwargs.pop('figsize', (4, 0.33*(extent[1]-extent[0])))
        else:
            fn = np.arange(Nf)
            xlabel = 'pseudofrequency [points]'
            figsize = kwargs.pop('figsize', (4, 13))

        _, fig = plot2d(BGNxx, extent=extent, figsize=figsize, title='Noise map',
                        ylabel=ylabel, xlabel = xlabel, vmin=vmin, vmax=vmax,
                        cmap=cmap, **kwargs)

        _, fig = plot2d(Sxx_out, extent=extent, figsize=figsize, title=title,
                        ylabel=ylabel, xlabel = xlabel, vmin=vmin, vmax=vmax,
                        cmap=cmap, **kwargs)

        fig2, (ax1, ax2) = plt.subplots(2, sharex=True)
        fig2.set_size_inches((5, 4))
        ax1, _ = plot1d(fn, np.mean(Sxx, axis=1), ax=ax1, legend='Original profile',
                        color='b',
                        xlabel='', ylabel= 'Amplitude [dB]', figtitle='')
        ax1, _ = plot1d(fn, np.mean(BGNxx, 1), ax=ax1, legend='Noise profile',
                        color='r',
                        xlabel='', ylabel='Amplitude [dB]', figtitle='')
        ax2, _ = plot1d(fn, np.mean(Sxx_out, axis=1), ax=ax2, color='k',
                        legend='Denoized profile',
                        xlabel=xlabel, ylabel= 'Amplitude [dB]', figtitle='')
        fig2.tight_layout()

        # SAVE FIGURE
        if savefig is not None:
            dpi = kwargs.pop('dpi', 96)
            dpi = kwargs.pop('dpi', 96)
            bbox_inches = kwargs.pop('bbox_inches', 'tight')
            format = kwargs.pop('format', 'png')
            savefilename = kwargs.pop(
                'savefilename', '_spectro_after_noise_subtraction')
            filename = savefig+savefilename+'.'+format
            print('\n''save figure : %s' % filename)
            fig.savefig(fname=filename, dpi=dpi, bbox_inches=bbox_inches,
                        format=format, **kwargs)

    return Sxx_out, noise_profile, BGNxx

# %%


def remove_background_along_axis(Sxx, mode='median', axis=1, N=25, N_bins=50,
                                  display=False, savefig=None, **kwargs):
    """ 
    Get the noisy profile along the defined axis and remove this profile from
    the spectrogram.

    Parameters 
    ---------- 
    Sxx : 2D numpy array  
        Original spectrogram (or image) 

    mode : str, optional, default is 'median'
        Select the mode to remove the noise
        Possible values for mode are :
        - 'ale' : Adaptative Level Equalization algorithm [Lamel & al. 1981]
        - 'median' : subtract the median value
        - 'mean' : subtract the mean value (DC)

    axis : integer, default is 1
        if matrix, estimate the mode for each row (axis=0) or each column (axis=1)

    N : int, default is 25
        length of window to compute the running mean of the noise profile

    N_bins : int (only for mode = "ale"), default is 50
        number of bins to compute the histogram 

    display : boolean, optional, default is False 
        Display the signal if True 

    savefig : string, optional, default is None 
        Root filename (with full path) is required to save the figures. Postfix 
        is added to the root filename. 

    kwargs, optional. This parameter is used by plt.plot and savefig functions 

        - savefilename : str, optional, default :'_spectro_after_noise_subtraction.png' 
            Postfix of the figure filename 

        - figsize : tuple of integers, optional, default: (4,10) 
            width, height in inches.   

        - title : string, optional, default : 'Spectrogram' 
            title of the figure 

        - xlabel : string, optional, default : 'Time [s]' 
            label of the horizontal axis 

        - ylabel : string, optional, default : 'Amplitude [AU]' 
            label of the vertical axis 

        - cmap : string or Colormap object, optional, default is 'gray' 
            See https://matplotlib.org/examples/color/colormaps_reference.html 
            in order to get all the  existing colormaps 
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic',  
            'viridis'... 

        - vmin, vmax : scalar, optional, default: None 
            `vmin` and `vmax` are used in conjunction with norm to normalize 
            luminance data.  Note if you pass a `norm` instance, your 
            settings for `vmin` and `vmax` will be ignored. 

        - extent : scalars (left, right, bottom, top), optional, default: None 
            The location, in data-coordinates, of the lower-left and 
            upper-right corners. If `None`, the image is positioned such that 
            the pixel centers fall on zero-based (row, column) indices. 

        - dpi : integer, optional, default is 96 
            Dot per inch.  
            For printed version, choose high dpi (i.e. dpi=300) => slow 
            For screen version, choose low dpi (i.e. dpi=96) => fast 

        - format : string, optional, default is 'png' 
            Format to save the figure 

        ... and more, see matplotlib    

    Returns 
    ------- 
    Sxx_out : 2d ndarray of scalar 
        Spectrogram after denoising 

    noise_profile : 1d ndarray of scalar
        Noise profile

    References
    ----------
    .. [1] Towsey, M., 2013. Noise Removal from Wave-forms and Spectrograms Derived from Natural Recordings of the Environment. Queensland University of Technology, Brisbane

    Examples
    --------
    Load audio recording and convert it into spectrogram

    >>> s, fs = maad.sound.load('../data/rock_savanna.wav')
    >>> Sxx,tn,fn,ext = maad.sound.spectrogram (s, fs)   

    Convert linear spectrogram into dB

    >>> Sxx_dB = maad.util.power2dB(Sxx) + 96

    Remove stationnary noise from the spectrogram with modes 'ale', 'median', and 'mean'.

    >>> Sxx_dB_noNoise_ale,_ = maad.sound.remove_background_along_axis(Sxx_dB, mode='ale')    
    >>> Sxx_dB_noNoise_med,_ = maad.sound.remove_background_along_axis(Sxx_dB, mode='median')
    >>> Sxx_dB_noNoise_mean,_ = maad.sound.remove_background_along_axis(Sxx_dB, mode='mean')

    Plot spectrograms

    >>> import matplotlib.pyplot as plt 
    >>> import numpy as np
    >>> fig, (ax1, ax2, ax3, ax4) = plt.subplots(4, 1)
    >>> maad.util.plot2d(Sxx_dB, ax=ax1, extent=ext, title='original', vmin=np.median(Sxx_dB), vmax=np.median(Sxx_dB)+40)
    >>> maad.util.plot2d(Sxx_dB_noNoise_ale, ax=ax2, extent=ext, title='Without stationary noise (mode = ''ale'')',vmin=np.median(Sxx_dB_noNoise_ale), vmax=np.median(Sxx_dB_noNoise_ale)+40)
    >>> maad.util.plot2d(Sxx_dB_noNoise_med, ax=ax3, extent=ext, title='Without stationary noise (mode = ''med'')',vmin=np.median(Sxx_dB_noNoise_med), vmax=np.median(Sxx_dB_noNoise_med)+40)
    >>> maad.util.plot2d(Sxx_dB_noNoise_mean, ax=ax4, extent=ext, title='Without stationary noise (mode = ''mean'')',vmin=np.median(Sxx_dB_noNoise_mean), vmax=np.median(Sxx_dB_noNoise_mean)+40)
    >>> fig.set_size_inches(8,10)
    >>> fig.tight_layout()   
    """

    # get the noise profile, N define the running mean size of the histogram
    # in case of mode='ale'
    noise_profile = get_unimode(Sxx, mode, axis, N=7, N_bins=N_bins)
    # smooth the profile by removing spurious thin peaks
    noise_profile = running_mean(noise_profile, N)
    # Remove horizontal noisy peaks profile (BGN_VerticalNoise is an estimation)
    # and negative value to zero
    if axis == 1:
        Sxx_out = Sxx - noise_profile[..., np.newaxis]
    elif axis == 0:
        Sxx_out = Sxx - noise_profile[np.newaxis, ...]

    # Set negative value to 0
    Sxx_out[Sxx_out < 0] = 0

    # Display
    if display:

        ylabel = kwargs.pop('ylabel', 'Frequency [Hz]')
        xlabel = kwargs.pop('xlabel', 'Time [sec]')
        title = kwargs.pop('title', 'Spectrogram without stationnary noise')
        cmap = kwargs.pop('cmap', 'gray')
        vmin = kwargs.pop('vmin', np.min(Sxx_out))
        vmax = kwargs.pop('vmax', np.max(Sxx_out))
        extent = kwargs.pop('extent', None)

        Nf, Nw = Sxx.shape

        if extent is not None:
            fn = np.arange(0, Nf)*(extent[3]-extent[2])/(Nf-1) + extent[2]
            xlabel = 'frequency [Hz]'
            figsize = kwargs.pop('figsize', (4, 0.33*(extent[1]-extent[0])))
        else:
            fn = np.arange(Nf)
            xlabel = 'pseudofrequency [points]'
            figsize = kwargs.pop('figsize', (4, 13))

        _, fig1 = plot2d(Sxx_out, extent=extent, figsize=figsize, title=title,
                         ylabel= ylabel, xlabel = xlabel, vmin=vmin, vmax=vmax,
                         cmap=cmap, **kwargs)

        fig2, (ax1, ax2) = plt.subplots(2, sharex=True)
        fig2.set_size_inches((5, 4))
        ax1, _ = plot1d(fn, mean_dB(Sxx, axis=axis), ax=ax1, legend='Original profile',
                        color='b',
                        xlabel='', ylabel= 'Amplitude [dB]', figtitle='')
        ax1, _ = plot1d(fn, noise_profile, ax=ax1, legend='Noise profile',
                        color='r',
                        xlabel='', ylabel='Amplitude [dB]', figtitle='')
        ax2, _ = plot1d(fn, mean_dB(Sxx_out, axis=axis), ax=ax2, color='k',
                        legend='Denoized profile',
                        xlabel=xlabel, ylabel= 'Amplitude [dB]', figtitle='')
        fig2.tight_layout()

        # SAVE FIGURE
        if savefig is not None:
            dpi = kwargs.pop('dpi', 96)
            dpi = kwargs.pop('dpi', 96)
            bbox_inches = kwargs.pop('bbox_inches', 'tight')
            format = kwargs.pop('format', 'png')
            savefilename = kwargs.pop(
                'savefilename', '_spectro_after_noise_subtraction')
            filename = savefig+savefilename+'.'+format
            print('\n''save figure : %s' % filename)
            fig1.savefig(fname=filename, dpi=dpi, bbox_inches=bbox_inches,
                         format=format, **kwargs)

    return Sxx_out, noise_profile

# %%


def median_equalizer(Sxx, display=False, savefig=None, **kwargs):
    """ 
    Remove background noise in spectrogram using median equalizer.

    Parameters 
    ---------- 
    Sxx : 2D numpy array  
        Original spectrogram (or image), !!! not in dB 

    display : boolean, optional, default is False 
        Display the signal if True 

    savefig : string, optional, default is None 
        Root filename (with full path) is required to save the figures. Postfix 
        is added to the root filename. 

    kwargs, optional. This parameter is used by plt.plot and savefig functions 

        - savefilename : str, optional, default :'_spectro_after_noise_subtraction.png' 
            Postfix of the figure filename 

        - figsize : tuple of integers, optional, default: (4,10) 
            width, height in inches.   

        - title : string, optional, default : 'Spectrogram' 
            title of the figure 

        - xlabel : string, optional, default : 'Time [s]' 
            label of the horizontal axis 

        - ylabel : string, optional, default : 'Amplitude [AU]' 
            label of the vertical axis 

        - cmap : string or Colormap object, optional, default is 'gray' 
            See https://matplotlib.org/examples/color/colormaps_reference.html 
            in order to get all the  existing colormaps 
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic',  
            'viridis'... 

        - vmin, vmax : scalar, optional, default: None 
            `vmin` and `vmax` are used in conjunction with norm to normalize 
            luminance data.  Note if you pass a `norm` instance, your 
            settings for `vmin` and `vmax` will be ignored. 

        - extent : scalars (left, right, bottom, top), optional, default: None 
            The location, in data-coordinates, of the lower-left and 
            upper-right corners. If `None`, the image is positioned such that 
            the pixel centers fall on zero-based (row, column) indices. 

        - dpi : integer, optional, default is 96 
            Dot per inch.  
            For printed version, choose high dpi (i.e. dpi=300) => slow 
            For screen version, choose low dpi (i.e. dpi=96) => fast 

        - format : string, optional, default is 'png' 
            Format to save the figure 

        ... and more, see matplotlib     

    Returns 
    ------- 
    Sxx_out : 2d ndarray of scalar 
        Spectrogram after denoising   

    References 
    ---------- 
    .. [1] This function has been proposed first by Carol BEDOYA <carol.bedoya@pg.canterbury.ac.nz> 
       Adapted by S. Haupert Oct 9, 2018 for Python 

    Examples
    --------

    Load audio recording and convert it into spectrogram

    >>> s, fs = maad.sound.load('../data/rock_savanna.wav')
    >>> Sxx,tn,fn,ext = maad.sound.spectrogram (s, fs)   

    Convert linear spectrogram into dB

    >>> Sxx_dB = maad.util.power2dB(Sxx) +96

    Remove stationnary noise from the spectrogram 

    >>> Sxx_noNoise = maad.sound.median_equalizer(Sxx)
    >>> Sxx_dB_noNoise =  maad.util.power2dB(Sxx_noNoise) 

    Plot both spectrograms

    >>> import matplotlib.pyplot as plt 
    >>> import numpy as np
    >>> fig, (ax1, ax2) = plt.subplots(2, 1)
    >>> maad.util.plot2d(Sxx_dB, ax=ax1, extent=ext, title='original', vmin=np.median(Sxx_dB), vmax=np.median(Sxx_dB)+40)
    >>> maad.util.plot2d(Sxx_dB_noNoise, ax=ax2, extent=ext, title='Without stationary noise',vmin=np.median(Sxx_dB_noNoise), vmax=np.median(Sxx_dB_noNoise)+40)
    >>> fig.set_size_inches(15,8)
    >>> fig.tight_layout() 

    """

    Sxx_out = (Sxx-np.median(Sxx, axis=1)[..., np.newaxis])

    # Numerator for normalization. Test if values of norm are <=0 and set them
    # to the highest value in Sxx. This will ensure that the result of the
    # normalization will be lower than 1
    norm = (np.median(Sxx, axis=1)-np.min(Sxx, axis=1))
    norm[norm <= 0] = Sxx.max()

    # normalization. Test if the numerator is 0
    Sxx_out = Sxx_out/norm[..., np.newaxis]

    # if the ratio is < 1, set the value to 1.
    # Values < 1 are noise and should not be less than 1.
    # When Sxx_out is converted into dB => log10(1) => 0
    Sxx_out[Sxx_out < 1] = 1

    # Display
    if display:
        ylabel = kwargs.pop('ylabel', 'Frequency [Hz]')
        xlabel = kwargs.pop('xlabel', 'Time [sec]')
        title = kwargs.pop('title', 'Spectrogram without stationnary noise')
        cmap = kwargs.pop('cmap', 'gray')
        extent = kwargs.pop('extent', None)

        if extent is not None:
            figsize = kwargs.pop('figsize', (4, 0.33*(extent[1]-extent[0])))
        else:
            figsize = kwargs.pop('figsize', (4, 13))

        # convert into dB
        Sxx_out_dB = power2dB(Sxx_out)

        vmin = kwargs.pop('vmin', 0)
        vmax = kwargs.pop('vmax', np.max(Sxx_out_dB))

        _, fig = plot2d(Sxx_out_dB, extent=extent, figsize=figsize, title=title,
                        ylabel=ylabel, xlabel = xlabel, vmin=vmin, vmax=vmax,
                        cmap=cmap, **kwargs)
        # SAVE FIGURE
        if savefig is not None:
            dpi = kwargs.pop('dpi', 96)
            dpi = kwargs.pop('dpi', 96)
            bbox_inches = kwargs.pop('bbox_inches', 'tight')
            format = kwargs.pop('format', 'png')
            savefilename = kwargs.pop(
                'savefilename', '_spectro_after_noise_subtraction')
            filename = savefig+savefilename+'.'+format
            print('\n''save figure : %s' % filename)
            fig.savefig(fname=filename, dpi=dpi, bbox_inches=bbox_inches,
                        format=format, **kwargs)

    return Sxx_out

#%%

def pcen(Sxx, gain=0.98, bias=2, power=0.5, b=0.025, eps=1e-6,
         display=False, savefig=False, **kwargs):
    """
       Per-Channel Energy Normalization (PCEN)

       This function normalizes a time-frequency representation Sxx by
       performing automatic gain control, followed by nonlinear compression [1]_
       This function was adapted from librosa PCEN function [2]_ to match linear
       spectrogram representation

       Parameters
       ----------
       Sxx : 2D numpy array
           Original spectrogram (or image)

       gain : scalar (>=0), optional, default is 0.98
           Gain factor. Typical values should be less than 1.

       bias : scalar (>=0), optional, default is 2
           Bias point of the nonlinear compression

       power : scalar (>=0), optional, default is 0.5
           Compression exponent. Typical values should be between 0 and 0.5.
           Smaller values of power result in stronger compression.
           At the limit power=0, polynomial compression becomes logarithmic.

       b : scalar between [0, 1], optional, default is 0.025
           The filter coefficient for the low-pass filter.

       eps : scalar (>0), optional, default is 1e-6
           A small constant used to ensure numerical stability of the filter.

       display : boolean, optional, default is False
           Display the signal if True

       savefig : string, optional, default is None
           Root filename (with full path) is required to save the figures. Postfix
           is added to the root filename.

       kwargs, optional. This parameter is used by plt.plot and savefig functions

           - savefilename : str, optional, default :'_spectro_after_noise_subtraction.png'
               Postfix of the figure filename

           - figsize : tuple of integers, optional, default: (4,10)
               width, height in inches.

           - title : string, optional, default : 'Spectrogram'
               title of the figure

           - xlabel : string, optional, default : 'Time [s]'
               label of the horizontal axis

           - ylabel : string, optional, default : 'Amplitude [AU]'
               label of the vertical axis

           - cmap : string or Colormap object, optional, default is 'gray'
               See https://matplotlib.org/examples/color/colormaps_reference.html
               in order to get all the  existing colormaps
               examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic',
               'viridis'...

           - vmin, vmax : scalar, optional, default: None
               `vmin` and `vmax` are used in conjunction with norm to normalize
               luminance data.  Note if you pass a `norm` instance, your
               settings for `vmin` and `vmax` will be ignored.

           - extent : scalars (left, right, bottom, top), optional, default: None
               The location, in data-coordinates, of the lower-left and
               upper-right corners. If `None`, the image is positioned such that
               the pixel centers fall on zero-based (row, column) indices.

           - dpi : integer, optional, default is 96
               Dot per inch.
               For printed version, choose high dpi (i.e. dpi=300) => slow
               For screen version, choose low dpi (i.e. dpi=96) => fast

           - format : string, optional, default is 'png'
               Format to save the figure

           ... and more, see matplotlib

       Returns
       -------
       Sxx_out : 2d ndarray of scalar
           Spectrogram after PCEN

       enhance_profile : 1d ndarray of scalar
           Enhance profile

       PCENxx : 2d ndarray of scalar
           enhanced map

       References
       ----------

       .. [1] Lostanlen, V., Salamon, J., McFee, B., Cartwright, M., Farnsworth, A., Kelling, S., and Bello, J. P. Per-Channel Energy Normalization: Why and How. IEEE Signal Processing Letters, 26(1), 39-43. `DOI: 10.1109/LSP.2018.2878620 <https://doi.org/10.1109/LSP.2018.2878620>`_ 
       .. [2] McFee, Brian, Colin Raffel, Dawen Liang, Daniel PW Ellis,Matt McVicar, Eric Battenberg, and Oriol Nieto. “librosa: Audio and music signal analysis in python.” In Proceedings of the 14th python in science conference, pp. 18-25. 2015. https://librosa.org/doc/main/generated/librosa.pcen.html

       """
    # PCEN
    # 1. smooth the spectrogram
    M = signal.lfilter([b], [1, b - 1], Sxx)
    # 2. take the inverse of smoothed spectrogram
    smooth = (eps + M)**(-gain)
    # 3. Nonlinear compression
    Sxx_out = (Sxx * smooth + bias)**power - bias**power

    # Enhance map (2d)
    PCENxx = Sxx - Sxx_out
    # Enhance profile along time axis (1d)
    enhance_profile = np.mean(PCENxx, 1)

    # Display
    if display:
         ylabel = kwargs.pop('ylabel', 'Frequency [Hz]')
         xlabel = kwargs.pop('xlabel', 'Time [sec]')
         title = kwargs.pop(
             'title', 'Spectrogram without stationnary noise after PCEN')
         cmap = kwargs.pop('cmap', 'gray')
         vmin = kwargs.pop('vmin', np.min(Sxx_out))
         vmax = kwargs.pop('vmax', np.max(Sxx_out))
         extent = kwargs.pop('extent', None)

         Nf, Nw = Sxx.shape

         if extent is not None:
             fn = np.arange(0, Nf)*(extent[3]-extent[2])/(Nf-1) + extent[2]
             xlabel = 'frequency [Hz]'
             figsize = kwargs.pop(
                 'figsize', (4, 0.33*(extent[1]-extent[0])))
         else:
             fn = np.arange(Nf)
             xlabel = 'pseudofrequency [points]'
             figsize = kwargs.pop('figsize', (4, 13))

         _, fig = plot2d(PCENxx, extent=extent, figsize=figsize, title='Enhance map',
                         ylabel=ylabel, xlabel=xlabel, vmin=vmin, vmax=vmax,
                         cmap=cmap, **kwargs)

         _, fig = plot2d(Sxx_out, extent=extent, figsize=figsize, title=title,
                         ylabel=ylabel, xlabel=xlabel, vmin=vmin, vmax=vmax,
                         cmap=cmap, **kwargs)

         fig2, (ax1, ax2) = plt.subplots(2, sharex=True)
         fig2.set_size_inches((5, 4))
         ax1, _ = plot1d(fn, np.mean(Sxx, axis=1), ax=ax1, legend='Original profile',
                         color='b',
                         xlabel='', ylabel='Amplitude [dB]', figtitle='')
         ax1, _ = plot1d(fn, np.mean(PCENxx, 1), ax=ax1, legend='Enhance profile',
                         color='r',
                         xlabel='', ylabel='Amplitude [dB]', figtitle='')
         ax2, _ = plot1d(fn, np.mean(Sxx_out, axis=1), ax=ax2, color='k',
                         legend='Profile after PCEN',
                         xlabel=xlabel, ylabel='Amplitude [dB]', figtitle='')
         fig2.tight_layout()

         # SAVE FIGURE
         if savefig is not None:
             dpi = kwargs.pop('dpi', 96)
             dpi = kwargs.pop('dpi', 96)
             bbox_inches = kwargs.pop('bbox_inches', 'tight')
             format = kwargs.pop('format', 'png')
             savefilename = kwargs.pop(
                 'savefilename', '_spectro_after_PCEN')
             filename = savefig+savefilename+'.'+format
             print('\n''save figure : %s' % filename)
             fig.savefig(fname=filename, dpi=dpi, bbox_inches=bbox_inches,
                         format=format, **kwargs)

    return Sxx_out, enhance_profile, PCENxx



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Functions to get audio metadata from files

Warning for Windows users. Due to problems using slashes and backslashes, Paths must be raw strings
instead of regular strings, to convert a regular string into a raw string simply add an r before the string.
for example:
    raw string:        r'C:/Users/Documents/Folder/SubFolder/file.wav'
"""

import wave
import glob
import os
import pandas as pd
from pathlib import Path
import numpy as np

#%%
def _ensure_directory(path_str):
    # Create a Path object from the input string
    path = Path(path_str)
    
    # Check if the path exists and is a directory
    if path.is_dir():
        return path
    elif path.exists():
        # If the path exists but is not a directory, raise an error
        raise ValueError(f"'{path_str}' exists but is not a directory.")
    else:
        # If the path does not exist, try adding a '/' at the end and check again
        path_with_slash = path / ''
        if path_with_slash.is_dir():
            return path_with_slash
        else:
            raise ValueError(f"'{path_str}' does not exist as a directory.")

#%%
def check_file_format(path_audio):
    """
    Check Wave file consistency. Check if WAVE format is correct and if file name 
    follows standard format. The standard format is SITENAME_DATE_TIME.WAV, with 
    DATE as YYYYMMDD and TIME as HHMMSS.

    Parameters
    ----------
    path_audio : str
        Location of audio filename.

    Raises
    ------
    File Not Found
        If file does not exist.

    Returns
    -------
    error : int
        0 if no error is found, 1 if WAVE format is incorrect and 2 if filename has no
        standard format.

    """
    basename = os.path.basename(path_audio)

    # Check Wave format:
    # try to open wav file, if error return only file name and null values on fields
    try:
        with wave.open(path_audio, 'rb') as f:
            _ = f.getparams()

    except FileNotFoundError as fnfe:
        raise fnfe

    except:
        error = 1
        return error

    # Check file name format:
    # 1. File name must have 3 fields separated by underscore '_'
    # 2. Second field (date) must have 8 characters
    # 3. Third field (time) must have 6 charcaters + 4 = 10 ('.WAV')
    if (len(basename.split('_')) != 3):
        error = 2
        return error

    else:
        date_str = basename.split('_')[1]
        time_str = basename.split('_')[2]
        if ((len(date_str) != 8) |  # date_str should have 8 characters
                (len(time_str) != 10) |  # time_str + '.wav' should have 10 characters
                (not (date_str.isnumeric())) |  # date_str should be numeric
                (not (time_str[0:-4].isnumeric()))  # time_str should be numeric
        ):
            error = 2
        else:
            error = 0

    return error

#%% 
def audio_header(path_audio):
    """
    Get audio header information from WAVE file. 
    Header information includes, sample rate, bit depth, number of channels, 
    number of samples, file size and duration.
    
    Parameters
    ----------
    path_audio : str
        Location of audio file.

    Returns
    -------
    metadata : dictionary
        header information.
    
    Examples
    --------
    >>> from maad import util
    >>> dic_metadata = util.audio_header('../data/spinetail.wav')
    >>> print(dic_metadata)
    {'path_audio': '../data/spinetail.wav', 'fname': 'spinetail.wav', 'sample_rate': 44100, 'channels': 1, 'bits': 16, 'samples': 861799, 'fsize': 1723642, 'length': 19.541927437641725}
    """
    basename = os.path.basename(path_audio)

    with wave.open(path_audio, 'rb') as f:
        meta = f.getparams()

    metadata = {'path_audio': path_audio,
                'fname': basename,
                'sample_rate': meta.framerate,
                'channels': meta.nchannels,
                'bits': meta.sampwidth * 8,
                'samples': meta.nframes,
                'fsize': os.path.getsize(path_audio),
                'length': meta.nframes / meta.framerate}
    return metadata

#%% 
def filename_info(path_audio, verbose =False):
    """
    Get information from filename when using standard format. The standard format is
    SITENAME_DATE_TIME.WAV, with DATE as YYYYMMDD and TIME as HHMMSS.

    Parameters
    ----------
    path_audio : str
        Location of audio file.

    Returns
    -------
    metadata : dictionary
        file name information.


    """
    if check_file_format(path_audio) == 0:
        basename = os.path.basename(path_audio)
        date = basename.split("_")[1]
        hour = basename.split("_")[2]
        date_fmt = date[0:4] + "-" + date[4:6] + "-" + date[6:8] + " " + hour[0:2] + ":" + hour[2:4] + ":" + hour[4:6]
        # structure data
        metadata = {'path_audio': path_audio,
                    'fname': basename,
                    'sensor_name': basename.split("_")[0],
                    'date': pd.to_datetime(date_fmt),
                    'time': basename.split("_")[2][0:6]}
    else:
        raise TypeError(
            'File name format not supported. The standard format must be SITENAME_DATE_TIME.WAV, with DATE as YYYYMMDD and TIME as HHMMSS.')
    return metadata

#%%
def get_metadata_file(path_audio, verbose=False):
    """
    Get metadata asociated with audio recordings in audio file. Metadata includes basic 
    information of the audio file format (sample rate, number of channels, bit depth and 
    file size), and date information from the filename. Note however, that this function 
    is intended for use only with audio files with a self-describing header.

    Parameters
    ----------
    path_audio : str
        Path to the audio file name.
    verbose : boolean, optional
        Display error messages. The default is False.

    Returns
    -------
    metadata : dictionary
        Dictionary with metadata.

    """
    path_audio = path_audio.replace('\\', '/')  # for compatibility with Windows

    basename = os.path.basename(path_audio)
    flag = check_file_format(path_audio)

    if flag == 1:  # unreadable audio file
        metadata = {'path_audio': path_audio,
                    'fname': basename,
                    'sample_rate': np.nan,
                    'channels': np.nan,
                    'bits': np.nan,
                    'samples': np.nan,
                    'fsize': np.nan,
                    'sensor_name': np.nan,
                    'date': np.nan,
                    'time': np.nan,
                    'length': np.nan}
        if verbose:
            print('Incorrect name or wave format. Return null values for: ', path_audio)
        return metadata

    elif flag == 2:  # filename format error (for passive acoustic monitoring only)
        info_header = audio_header(path_audio)
        metadata = {'path_audio': path_audio,
                    'fname': info_header['fname'],
                    'sample_rate': info_header['sample_rate'],
                    'channels': info_header['channels'],
                    'bits': info_header['bits'],
                    'samples': info_header['samples'],
                    'length': info_header['length'],
                    'fsize': info_header['fsize'],
                    'sensor_name': np.nan,
                    'date': np.nan,
                    'time': np.nan}
        return metadata

    else:  # No error,execute function
        info_header = audio_header(path_audio)
        info_fname = filename_info(path_audio)
        metadata = {'path_audio': path_audio,
                    'fname': info_header['fname'],
                    'sample_rate': info_header['sample_rate'],
                    'channels': info_header['channels'],
                    'bits': info_header['bits'],
                    'samples': info_header['samples'],
                    'length': info_header['length'],
                    'fsize': info_header['fsize'],
                    'sensor_name': info_fname['sensor_name'],
                    'date': info_fname['date'],
                    'time': info_fname['time']}
    return metadata

# %%
def get_metadata_dir(path_dir, verbose=False):
    """
    Get metadata asociated with audio recordings in a directory. Metadata includes basic 
    information of the audio file format (sample rate, number of channels, bit depth and 
    file size), and date information from the filename. Note however, that this function 
    is intended for use only with audio files with a self-describing header.

    Parameters
    ----------
    path_dir : str
        Path of either a directory or a file. it will select all wav files in the parent folder
        (of either the file or directory in path_dir). The search for file is performed recursively.
    verbose : boolean, optional
        Output file progress. The default is False.

    Returns
    -------
    df_metadata : pandas.DataFrame
        Dataframe with metadata, files as rows and metadata as columns.
    
    See Also
    --------
    maad.util.get_metadata_file, maad.util.audio_header, maad.util.filename_info
    
    Examples
    --------
    >>> from maad import util
    >>> df_metadata = util.get_metadata_dir('../data/indices/')
    """
    # Verify that input is a directory
    path_dir = _ensure_directory(path_dir)

    # List all files recursively and select only wav files.
    flist_wav = glob.glob(f"{path_dir}/**/*.[Ww][Aa][Vv]", recursive=True)

    # Get metadata for each file
    df_metadata = pd.DataFrame()
    for count, file in enumerate(flist_wav):
        if verbose:
            print(f'{count +1} / {len(flist_wav)} : {os.path.basename(file)}', end='\r')

        data = get_metadata_file(file, verbose)
        df_metadata = pd.concat([df_metadata, pd.DataFrame.from_records([data])])

    df_metadata.reset_index(drop=True, inplace=True)
    
    return df_metadata



#%% Examples of use
"""
Test functions 


# 1. Read metadata from a valid file with correct name format -> read ok
path_audio = './test_data/S4A03895_20190522_000000.wav'
get_metadata_file(path_audio)
audio_header(path_audio)
filename_info(path_audio)
check_file_format(path_audio)  # should be 0

# 2. Read metadata from a valid file with incorrect name format -> null output
path_audio = './test_data/spinetail_20220219_30222L.wav'
get_metadata_file(path_audio)
audio_header(path_audio)
filename_info(path_audio)
check_file_format(path_audio) # should be error=2

# 3. Read metadata from a wav file with incorrect format and correct name format -> null output
path_audio = './test_data/NOHEADER_20190522_000000.wav'
get_metadata_file(path_audio)
audio_header(path_audio)
filename_info(path_audio)
check_file_format(path_audio)  # should be error=1

# 4. Read metadata from a file with no header and incorrect name format -> null output
path_audio = './test_data/NOHEADER_BAD_FNAME_FORMAT.wav'
get_metadata_file(path_audio)
audio_header(path_audio)
filename_info(path_audio)
check_file_format(path_audio)  # should be error=1

"""





# -*- coding: utf-8 -*-
""" 
Sound processing
================

The module ``sound`` is an ensemble of functions to load and preprocess audio signals.

Input and output
-----------------
.. autosummary::
    :toctree: generated/

    load
    load_url
    load_spectrogram
    write
    
Preprocess audio
-----------------
.. autosummary::
    :toctree: generated/

    fir_filter
    sinc
    smooth
    select_bandwidth
    pcen
    remove_background
    remove_background_morpho
    remove_background_along_axis
    median_equalizer
    wave2frames
   
Transform audio
---------------
.. autosummary::
    :toctree: generated/
    
    spectrogram
    avg_power_spectro
    avg_amplitude_spectro
    linear_to_octave
    envelope
    spectrum
    resample
    trim
    normalize
    gain

Metrics
-------
.. autosummary::
    :toctree: generated/
    
    temporal_snr
    spectral_snr
    sharpness

"""

from .input_output import (load,
                           load_url,
                           load_spectrogram,
                           write)

from .filter import (select_bandwidth,
                     fir_filter,
                     sinc,
                     smooth)

from .spectral_subtraction import (pcen,
                                   remove_background,
                                   remove_background_morpho,
                                   remove_background_along_axis,
                                   median_equalizer)

from .trim_func import wave2frames

from .transform import (envelope,
                        spectrum,
                        resample,
                        trim,
                        normalize,
                        gain)

from .spectro_func import (spectrogram,
                           avg_power_spectro,
                           avg_amplitude_spectro,
                           linear_to_octave)
                          
from .metrics import (temporal_snr,
                      spectral_snr,
                      sharpness)

__all__ = [
        # io.py
        'load',
        'load_url',
        'load_spectrogram',
        'write',
        # filter.py
        'select_bandwidth',
        'fir_filter',
        'sinc',
        'smooth',
        # spectral_subtraction.py
        'pcen',
        'remove_background', 
        'remove_background_morpho',
        'remove_background_along_axis',
        'median_equalizer',
        # trim_func.py
        'wave2frames',
        # transform.py
        'envelope',
        'spectrum',
        'resample',
        'trim',
        'normalize',
        'gain',
        # spectro_func.py
        'spectrogram',
        'avg_power_spectro',
        'avg_amplitude_spectro',
        'linear_to_octave',
        # metrics.py
        'temporal_snr',
        'spectral_snr',
        'sharpness']


#!/usr/bin/env python
""" 
Collection of functions transform an audio signal into spectrogram and 
manipulate spectrograms.
"""
#
# Authors:  Juan Sebastian ULLOA <lisofomia@gmail.com>
#           Sylvain HAUPERT <sylvain.haupert@mnhn.fr>
#
# License: New BSD License

# =============================================================================
# Load the modules
# =============================================================================
# Import external modules
import numpy as np
import scipy as sp

# Import internal modules
from maad.util import (plot1d, plot2d, crop_image, power2dB, amplitude2dB)

# %%
# =============================================================================
# public functions
# =============================================================================


def spectrogram(x, fs, window='hann', nperseg=1024, noverlap=None,
                flims=None, tlims=None,
                mode='psd',
                verbose=False, display=False,
                savefig=None, **kwargs):
    """
    Compute a spectrogram using the short-time Fourier transform from an audio signal.

    The function can compute diferent outputs according to the parameter 'mode':
        - power (mode='psd')
        - amplitude (mode = 'amplitude') => sqrt(power)
        - complex with real and imaginary parts (mode = 'complex')

    Parameters
    ----------
    x : 1d ndarray
        Vector containing the sound waveform 
    fs : int
        The sampling frequency in Hz 
    window : str or tuple or array_like, optional, default to 'hann'
        Desired window to use. If window is a string or tuple, it is passed to 
        get_window to generate the window values, which are DFT-even by default. 
        See get_window for a list of windows and required parameters. 
        If window is array_like it will be used directly as the window and its length must be nperseg. 
    nperseg : int, optional. Defaults to 1024.
        Length of the segment used to compute the FFT. No zero padding. 
        For fast calculation, it's better to use a number that is a power 2. 
        This parameter sets the resolution in frequency as the spectrogram will
        contains nperseg/2 frequency bins between 0Hz-(fs/2)Hz, with a resolution
        df = fs/nperseg
        It sets also the time slot (dt) of each frequency frames : dt = nperseg/fs
        The higher is the number, the lower is the resolution in time (dt) 
        but better is the resolution in frequency (df).  
    noverlap : int, optional. Defaults to None.
        Number of points to overlap between segments. 
        If None, noverlap = nperseg // 2. 
    mode : str, optional. Default is 'psd'
        Choose the output between 
        - 'psd' : Power Spectral Density 
        - 'amplitude' : module of the stft (sqrt(psd))
        - 'complex' : real and imaginary part of the stft  
    flims, tlims : list of 2 scalars [min, max], optional, default is None
        flims corresponds to the min and max boundary frequency values
        tlims corresponds to the min and max boundary time values  
    verbose : boolean, optional, default is False
        print messages into the consol or terminal if verbose is True   
    display : boolean, optional, default is False
        Display the signal if True
    savefig : string, optional, default is None
        Root filename (with full path) is required to save the figures. Postfix
        is added to the root filename.        
    **kwargs, optional. This parameter is used by plt.plot and savefig functions
        - savefilename : str, optional, default :'_filt_audiogram.png'
            Postfix of the figure filename
        - db_range : scalar, optional, default : 100
            if db_range is a number, anything lower than -db_range is set to 
            -db_range and anything larger than 0 is set to 0
        - figsize : tuple of integers, optional, default: (4,10)
            width, height in inches.  
        - title : string, optional, default : 'Spectrogram'
            title of the figure
        - xlabel : string, optional, default : 'Time [s]'
            label of the horizontal axis
        - ylabel : string, optional, default : 'Amplitude [AU]'
        - cmap : string or Colormap object, optional, default is 'gray'
            See https://matplotlib.org/examples/color/colormaps_reference.html
            in order to get all the  existing colormaps
            examples: 'hsv', 'hot', 'bone', 'tab20c', 'jet', 'seismic', 
            'viridis'...
        - vmin, vmax : scalar, optional, default: None
            `vmin` and `vmax` are used in conjunction with norm to normalize
            luminance data.  Note if you pass a `norm` instance, your
            settings for `vmin` and `vmax` will be ignored.

        - extent : list of scalars [left, right, bottom, top], optional, default: None
            The location, in data-coordinates, of the lower-left and
            upper-right corners. If `None`, the image is positioned such that
            the pixel centers fall on zero-based (row, column) indices.
        - dpi : integer, optional, default is 96
            Dot per inch. 
            For printed version, choose high dpi (i.e. dpi=300) => slow
            For screen version, choose low dpi (i.e. dpi=96) => fast
        - format : string, optional, default is 'png'
            Format to save the figure

        ... and more, see matplotlib  

    Returns
    -------
    Sxx : 2d ndarray of floats
        Spectrogram : Matrix containing K frames with N/2 frequency bins, 
        K*N <= length (wave)
        Sxx unit is power => Sxx_power if mode is 'psd'
        Sxx unit is amplitude => Sxx_ampli if mode is 'amplitude' or 'complex'
    tn : 1d ndarray of floats
        time vector (horizontal x-axis)    
    fn : 1d ndarray of floats
        Frequency vector (vertical y-axis)    
    extent : list of scalars [left, right, bottom, top]
        The location, in data-coordinates, of the lower-left and
        upper-right corners. 

    Notes
    -----
    This function takes care of the energy conservation which is crucial 
    when working with sound pressure level (dB SPL)

    Examples
    --------
    >>> s,fs = maad.sound.load("../data/rock_savanna.wav")

    Compute energy of signal s

    >>> E1 = sum(s**2)
    >>> maad.util.power2dB(E1)
    44.861029507805256

    Compute the spectrogram with 'psd' output (if N<4096, the energy is lost)

    >>> N = 4096
    >>> Sxx_power,tn,fn,ext = maad.sound.spectrogram (s, fs, nperseg=N, noverlap=N//2, mode = 'psd')   

    Display Power Spectrogram

    >>> Sxx_dB = maad.util.power2dB(Sxx_power) # convert into dB
    >>> fig_kwargs = {'vmax': Sxx_dB.max(),
                      'vmin':-70,
                      'extent':ext,
                      'figsize':(4,13),
                      'title':'Power spectrogram density (PSD)',
                      'xlabel':'Time [sec]',
                      'ylabel':'Frequency [Hz]',
                      }
    >>> fig, ax = maad.util.plot2d(Sxx_dB,**fig_kwargs)     

    Compute mean power spectrogram

    >>> S_power_mean = maad.sound.avg_power_spectro(Sxx_power)

    energy => power x time

    >>> E2 = sum(S_power_mean*len(s)) 
    >>> maad.util.power2dB(E2)
    44.93083283875093

    Compute the spectrogram with 'amplitude' output

    >>> Sxx_ampli,tn,fn,_ = maad.sound.spectrogram (s, fs, nperseg=N, noverlap=N//2, mode='amplitude')  

    For energy conservation => convert Sxx_ampli (amplitude) into power before doing the average.

    >>> S_ampli_mean = maad.sound.avg_amplitude_spectro(Sxx_ampli)
    >>> S_power_mean = S_ampli_mean**2

    energy => power x time

    >>> E3 = sum(S_power_mean*len(s)) 
    >>> maad.util.power2dB(E3)
    44.93083283875093

    """
    
    # Get the argument detrend. By default is "constant" but some reasons (ADI
    # and AEI index from soundecology), detrend should be None
    detrend = kwargs.pop("detrend", "constant")

    # Test if noverlap is None. By default, noverlap is half the length of the fft
    if noverlap is None:
        noverlap = nperseg // 2

    # compute the number of frames
    K = len(x)//(nperseg-noverlap)-1

    # compute spectrogram
    fn, tn, Sxx_complex = sp.signal.spectrogram(x, fs, window=window,
                                                nperseg=nperseg, noverlap=noverlap,
                                                nfft=nperseg,
                                                mode='complex',
                                                detrend=detrend,
                                                scaling='density', axis=-1)

    # Mutliply by the frequency resolution step (fs/nperseg) to get the power
    # so multiply by the sqrt((fs/nperseg)) to get the amplitude
    # Also multiply by sqrt(2) in order to compensate that only the postive frequencies are kept
    # complex
    Sxx_complex = Sxx_complex*np.sqrt(2*(fs/nperseg))
    # magnitude
    Sxx = abs(Sxx_complex)
    # power
    PSDxx = Sxx**2

    if mode == 'complex':
        Sxx_out = Sxx_complex

    if mode == 'amplitude':
        Sxx_out = Sxx

    if mode == 'psd':
        Sxx_out = PSDxx

    # test if the last frames are computed on a whole time frame.
    # if note => remove these frames
    if PSDxx.shape[1] > K:
        sup = Sxx_out.shape[1] - K
        Sxx_out = Sxx_out[:, :-sup]
        tn = tn[:-sup]

    # Remove the last frequency bin in order to obtain nperseg/2 frequency bins
    # instead of nperseg/2 + 1
    Sxx_out = Sxx_out[:-1, :]
    fn = fn[:-1]

    if verbose:
        print('spectrogram dimension Nx=%d Ny=%d' %
              (Sxx_complex.shape[0], Sxx_complex.shape[1]))

    # Crop the image in order to analyzed only a portion of it
    if (flims or tlims) is not None:
        if verbose:
            print('Crop the spectrogram along time axis and frequency axis')
        Sxx_out, tn, fn = crop_image(Sxx_out, tn, fn, flims, tlims)

    if verbose:
        print('max value of the spectrogram %.5f' % Sxx_out.max())

    # Extent
    extent = [tn[0], tn[-1], fn[0], fn[-1]]
    # dt and df resolution
    dt = tn[1]-tn[0]
    df = fn[1]-fn[0]

    if verbose == True:
        print("*************************************************************")
        print("   Time resolution dt=%.2fs | Frequency resolution df=%.2fHz "
              % (dt, df))
        print("*************************************************************")

    # Display
    if display:

        ylabel = kwargs.pop('ylabel', 'Frequency [Hz]')
        xlabel = kwargs.pop('xlabel', 'Time [sec]')
        title = kwargs.pop('title', 'Spectrogram')
        cmap = kwargs.pop('cmap', 'gray')
        figsize = kwargs.pop('figsize', (4, 0.33*(extent[1]-extent[0])))
        db_range = kwargs.pop('db_range', 96)

        # convert into dB
        if mode == 'psd':
            Sxx_disp = power2dB(Sxx_out, db_range=db_range)
        if mode == 'amplitude':
            Sxx_disp = amplitude2dB(Sxx_out, db_range=db_range)
        if mode == 'complex':
            Sxx_disp = amplitude2dB(Sxx_out, db_range=db_range)

        vmin = kwargs.pop('vmin', -db_range)
        vmax = kwargs.pop('vmax', Sxx_disp.max())

        _, fig = plot2d(Sxx_disp, extent=extent, figsize=figsize, title=title,
                        ylabel=ylabel, xlabel=xlabel, vmin=vmin, vmax=vmax,
                        cmap=cmap, **kwargs)
        # SAVE FIGURE
        if savefig is not None:
            dpi = kwargs.pop('dpi', 96)
            bbox_inches = kwargs.pop('bbox_inches', 'tight')
            format = kwargs.pop('format', 'png')
            savefilename = kwargs.pop('savefilename', '_spectrogram')
            filename = savefig+savefilename+'.'+format
            print('\n''save figure : %s' % filename)
            fig.savefig(fname=filename, dpi=dpi, bbox_inches=bbox_inches,
                        format=format, **kwargs)

    return Sxx_out, tn, fn, extent

# %%


def linear_to_octave(X, fn, thirdOctave=True, display=False, **kwargs):
    """
    Transform a linear spectrum (1d) or Spectrogram (2d into octave or 1/3 octave
    spectrum (1d) or Spectrogram (2d).

    Our advice is to work with PSD (amplitude²) for energy conservation.

    Parameters
    ----------
    X : ndarray of floats
        Linear spectrum (1d) or Spectrogram (2d). 
        Work with PSD to be consistent with energy conservation
    fn : 1d ndarray of floats
        Frequency vector of the linear spectrum/spectrogram
    thirdOctave : Boolean, default is True
        choose between Octave or thirdOctave frequency resolution
    display : boolean, default is False
        Display the octave spectrum/spectrogram
    ** kwargs : optional. This parameter is used by plt.plot    

    Returns
    -------
    X_octave : ndarray of floats
        Octave or 1/3 octave Spectrum (1d) or Spectrogram (2d)
    bin_octave : vector of floats
        New frequency vector (octave or 1/3 octave frequency repartition)

    Examples
    --------
    >>> w, fs = maad.sound.load('../data/rock_savanna.wav') 
    >>> Sxx_power,tn,fn, ext = maad.sound.spectrogram (w, fs, nperseg=8192)
    >>> maad.sound.linear_to_octave(Sxx_power, fn, display=True, extent=ext, vmin=-50)
    """

    # define the third octave or octave frequency vector in Hz.
    if thirdOctave:
        bin_octave = np.array([16, 20, 25, 31.5, 40, 50, 63, 80, 100, 125, 160, 200, 250, 315,
                               400, 500, 630, 800, 1000, 1250, 1600, 2000, 2500, 3150, 4000,
                               5000, 6300, 8000, 10000, 12500, 16000, 20000])  # third octave band.
    else:
        bin_octave = np.array(
            [16, 31, 63, 125, 250, 500, 1000, 2000, 4000, 8000, 16000])  # octave

    # get the corresponding octave from fn
    bin_octave = bin_octave[(bin_octave >= np.min(fn))
                            & (bin_octave <= np.max(fn))]

    # Bins limit
    bin_octave_low = bin_octave/(2**0.1666666)
    bin_octave_up = bin_octave*(2**0.1666666)

    # select the indices corresponding to the frequency bins range
    X_octave = []
    for ii in np.arange(len(bin_octave)):
        ind = (fn >= bin_octave_low[ii]) & (fn <= bin_octave_up[ii])
        X_octave.append(np.sum(X[ind, ], axis=0))

    X_octave = np.asarray(X_octave)

    if display:
        X_octave_dB = power2dB(X_octave)
        if np.ndim(X_octave_dB) == 2:
            extent = kwargs.pop('extent', None)
            if extent is not None:
                xlabel = 'Time [sec]'
                duration = extent[1] - extent[0]
                deltaT = int(duration/10)
                xticks = (np.floor(np.arange(0, X.shape[1], X.shape[1]/10)*10)/10,
                          np.floor(np.arange(extent[0], extent[1], deltaT)*10)/10)
                figsize = (4, 0.33*(extent[1]-extent[0]))
            else:
                xlabel = 'pseudoTime [points]'
                xticks = np.arange(0, X.shape[1], 100),
                figsize = (4, 13)

            fig_kwargs = {'vmax': kwargs.pop('vmax', np.max(X_octave_dB)),
                          'vmin': kwargs.pop('vmin', np.min(X_octave_dB)),
                          'xticks': xticks,
                          'figsize': kwargs.pop('figsize', figsize),
                          'yticks': (np.arange(len(bin_octave)), bin_octave),
                          'title': 'Octave Spectrogram',
                          'xlabel': xlabel,
                          'ylabel': 'Frequency [Hz]',
                          }
            plot2d(X_octave_dB, **fig_kwargs)
        elif np.ndim(X_octave_dB) == 1:
            fig_kwargs = {
                'title': 'Octave Spectrum',
                'xlabel': kwargs.pop('xlabel', 'Frequency [Hz]'),
                'ylabel': kwargs.pop('ylabel', 'Amplitude [dB]'),
            }
            plot1d(bin_octave, X_octave_dB, **fig_kwargs)

    return X_octave, bin_octave

# %%


def avg_power_spectro(Sxx_power):
    """
    Computes the average of a power spectrogram along the time axis.

    Parameters
    ----------
    Sxx_power : 2d ndarray of floats
        Power spectrogram [Matrix]

    Returns
    -------
    S_power_mean : 1d ndarray of floats
        Power spectrum [Vector]

    See Also
    --------
    avg_amplitude_spectro

    Examples
    --------
    >>> w, fs = maad.sound.load('../data/rock_savanna.wav') 
    >>> Sxx_power,tn, fn, ext = maad.sound.spectrogram (w, fs)

    Compute mean power spectrogram

    >>> S_power_mean = maad.sound.avg_power_spectro(Sxx_power)
    >>> S_power_mean
    array([2.91313297e-04, 9.16802665e-04, 2.89179556e-04, 1.04767281e-04,
       6.66154492e-05, 4.59636926e-05, 3.59688131e-05, 3.30869371e-05,
       ...


    """
    S_power_mean = np.mean(Sxx_power, axis=1)

    return S_power_mean

# %%

def avg_amplitude_spectro(Sxx_ampli):
    """
    Computes the average of an amplitude spectrogram along the time axis.

    Parameters
    ----------
    Sxx_ampli : 2d ndarray of floats
        Amplitude spectrogram [Matrix]

    Returns
    -------
    S_ampli_mean : 1d ndarray of floats
        Amplitude spectrum [Vector]

    See Also
    --------
    avg_power_spectro

    Examples
    --------
    >>> w, fs = maad.sound.load('../data/rock_savanna.wav') 
    >>> Sxx_amplitude,tn, fn, ext = maad.sound.spectrogram (w, fs, mode="amplitude")

    Compute mean amplitude spectrogram

    >>> S_amplitude_mean = maad.sound.avg_amplitude_spectro(Sxx_amplitude)
    >>> S_amplitude_mean
    array([0.0170679 , 0.03027875, 0.01700528, 0.01023559, 0.00816183,
       0.00677965, 0.0059974 , 0.00575212, 0.00700752, 0.00926279,
       ...

    """

    # average the amplitude spectrogram taking the PSD for energy conservation
    S_ampli_mean = np.sqrt(np.mean(Sxx_ampli**2, axis=1))

    return S_ampli_mean
